package org.expeditee.actions;

import java.awt.Color;
import java.awt.Point;
import java.awt.event.InputEvent;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import org.expeditee.gui.AttributeUtils;
import org.expeditee.gui.DisplayIO;
import org.expeditee.gui.Frame;
import org.expeditee.gui.FrameGraphics;
import org.expeditee.gui.FrameIO;
import org.expeditee.gui.FrameMouseActions;
import org.expeditee.gui.FrameUtils;
import org.expeditee.io.Conversion;
import org.expeditee.items.Dot;
import org.expeditee.items.Item;
import org.expeditee.items.Line;
import org.expeditee.items.Text;
import org.expeditee.simple.BelowMinParametreCountException;
import org.expeditee.simple.Context;
import org.expeditee.simple.IncorrectParametreCountException;
import org.expeditee.simple.IncorrectTypeException;
import org.expeditee.simple.Pointers;
import org.expeditee.simple.Primitives;
import org.expeditee.simple.SBoolean;
import org.expeditee.simple.SCharacter;
import org.expeditee.simple.SInteger;
import org.expeditee.simple.SPointer;
import org.expeditee.simple.SPrimitive;
import org.expeditee.simple.SReal;
import org.expeditee.simple.SString;
import org.expeditee.simple.SVariable;
import org.expeditee.simple.UnitTestFailedException;
import org.expeditee.stats.AgentStats;

public class Simple {

	private static final String DEFAULT_STRING = "$s.";

	private static final String DEFAULT_BOOLEAN = "$b.";

	private static final String DEFAULT_CHAR = "$c.";

	private static final String DEFAULT_INTEGER = "$i.";

	private static final String DEFAULT_REAL = "$r.";

	private static final String DEFAULT_ITEM = "$ip.";

	private static final String DEFAULT_FRAME = "$fp.";

	private static final String EXIT_TEXT = "exitall";

	private static enum Status {
		Exit, OK, Break, Continue, Return, TrueIf, FalseIf;
	};

	private static final String BREAK_TEXT = "exitrepeat";

	private static final String CONTINUE_TEXT = "nextrepeat";

	private static final String RETURN_TEXT = "return";

	private static final String LOOP_TEXT = "repeat";

	private static final String TOKEN_SEPARATOR = " +";

	private static final String RUN_FRAME_ACTION = "RunFrame";

	/**
	 * Keeps track of how many simple programs are running. Used to check if
	 * simple should read in keyboard input. Or if the keyboard input should be
	 * handled normally by Expeditee.
	 */
	private static int _programsRunning = 0;

	public static void ProgramFinished() {
		_programsRunning--;
	}

	private static LinkedList<Character> _KeyStrokes = new LinkedList<Character>();

	public static void KeyStroke(char c) {
		_KeyStrokes.add(c);
	}

	public static boolean isProgramRunning() {
		return _programsRunning > 0;
	}

	public static void NewSimpleTest() {
		Frame newSimpleTest = FrameIO.CreateFrame(DisplayIO.getCurrentFrame()
				.getFramesetName(), "Test", null);
		List<String> actions = new ArrayList<String>();
		actions.add(RUN_FRAME_ACTION);
		newSimpleTest.getTitle().setAction(actions);
		FrameUtils.DisplayFrame(newSimpleTest, true);
		FrameGraphics.DisplayMessage("New test created");
	}

	public static void NextTest() {
		Frame next = DisplayIO.getCurrentFrame();
		do {
			next = FrameIO.LoadNext(next);
		} while (next != null
				&& (next.getTitle() == null || !RUN_FRAME_ACTION
						.equalsIgnoreCase(next.getTitle().getFirstAction())));
		FrameUtils.DisplayFrame(next, true);
	}

	public static void PreviousTest() {
		Frame prev = DisplayIO.getCurrentFrame();
		do {
			prev = FrameIO.LoadPrevious(prev);
		} while (prev != null
				&& (prev.getTitle() == null || !RUN_FRAME_ACTION
						.equalsIgnoreCase(prev.getTitle().getFirstAction())));

		FrameUtils.DisplayFrame(prev, true);
	}

	public static void LastTest() {
		Frame next = FrameIO.LoadLast();
		Frame lastTest = null;
		do {
			// check if its a test frame
			if (next != null
					&& next.getTitle() != null
					&& RUN_FRAME_ACTION.equalsIgnoreCase(next.getTitle()
							.getFirstAction())) {
				lastTest = next;
				break;
			}

			next = FrameIO.LoadPrevious(next);
		} while (next != null);

		FrameUtils.DisplayFrame(lastTest, true);
	}

	public static void RunSimpleTests(String frameset) {
		RunSimpleTests(frameset, false);
	}

	public static void RunSimpleTestsVerbose(String frameset) {
		RunSimpleTests(frameset, true);
	}

	public static void RunSimpleTests(String frameset, boolean verbose) {
		int testsPassed = 0;
		int testsFailed = 0;

		FrameIO.SaveFrame(DisplayIO.getCurrentFrame(), false);

		FrameGraphics.DisplayMessage("Starting test suite: " + frameset,
				Color.CYAN);

		// Get the next number in the inf file for the frameset
		int lastFrameNo = FrameIO.getLastNumber(frameset);

		// Loop through all the valid frames in the frameset
		for (int i = 1; i <= lastFrameNo; i++) {
			String nextFrameName = frameset + i;
			Frame nextFrame = FrameIO.LoadFrame(nextFrameName);
			if (nextFrame != null) {
				// Run the frames with the RunFrame action on the title
				Text title = nextFrame.getTitle().copy();
				List<String> actions = title.getAction();
				if (actions != null && !title.isAnnotation()) {
					if (actions.get(0).toLowerCase().equals("runframe")) {
						try {
							title.setLink(nextFrameName);
							// TODO add the ability to run a setup frame
							// which sets up variables to be used in all
							// tests
							AgentStats.reset();
							_KeyStrokes.clear();
							_programsRunning++;
							RunFrameAndReportError(title, new Context());
							_programsRunning--;
							if (verbose)
								FrameGraphics.DisplayMessage("Test passed: "
										+ title.toString(), Item.GREEN);
							testsPassed++;
						} catch (Exception e) {
							testsFailed++;
							_programsRunning--;
							// Print out the reason for failed tests
							FrameGraphics.LinkedErrorMessage(e.getMessage());
						}
					}
				}
			}
		}

		// Report the number of test passed and failed
		FrameGraphics.DisplayMessage("Total tests: "
				+ (testsPassed + testsFailed), Color.CYAN);
		if (testsPassed > 0)
			FrameGraphics.DisplayMessage("Passed: " + testsPassed, Item.GREEN);
		if (testsFailed > 0)
			FrameGraphics.DisplayMessage("Failed: " + testsFailed, Color.RED);
	}

	public static void RunFrame(Item current) {
		try {
			FrameIO.SaveFrame(DisplayIO.getCurrentFrame(), true);

			// an item without a link signals to run the current frame
			if (current.getLink() == null) {
				current = current.copy();
				current.setLink(DisplayIO.getCurrentFrame().getFrameName());
			}

			_KeyStrokes.clear();

			_programsRunning++;
			Thread t = new Thread(current);
			t.setPriority(Thread.MIN_PRIORITY);
			t.start();
			// _programsRunning--;
		} catch (Exception e) {
			// _programsRunning--;
		}
	}

	private static void FlagError(Item item) {
		item.setColor(Color.CYAN);
		FrameUtils.DisplayFrame(item.getParent().getFrameName(), true);
		FrameIO.SaveFrame(item.getParent());
	}

	/**
	 * Runs a simple code begining on a frame linked to by the specified item
	 * parametre.
	 * 
	 * @param current
	 *            the item that is linked to the frame to be run.
	 */
	public static Status RunFrameAndReportError(Item current, Context context)
			throws Exception {
		// the item must link to a frame
		if (current.getLink() == null) {
			throw new Exception("Could not run unlinked item: "
					+ current.toString());
		}

		Frame child = FrameIO.LoadFrame(current.getAbsoluteLink());
		AgentStats.FrameExecuted();

		// if the frame could not be loaded
		if (child == null) {
			throw new Exception("Could not load item link: "
					+ current.toString());
		}

		// loop through non-title, non-name, non-annotation text items
		List<Item> body = child.getItems();
		Text item = null;
		Status lastItemStatus = Status.OK;
		for (Item i : body)
			// Ignore items that arent statements
			if (isStatement(i)) {
				item = (Text) i;
				AgentStats.ItemExecuted();
				try {

					lastItemStatus = RunItem(item, context, lastItemStatus);

					if (lastItemStatus != Status.OK) {
						if (lastItemStatus != Status.TrueIf
								&& lastItemStatus != Status.FalseIf) {
							return lastItemStatus;
						}
					}
				} catch (ArrayIndexOutOfBoundsException e) {
					FlagError(item);
					throw new Exception("Too few parametres: "
							+ item.toString());
				} catch (NullPointerException e) {
					FlagError(item);
					throw new Exception("Null pointer exception: "
							+ item.toString());
				} catch (RuntimeException e) {
					FlagError(item);
					throw new Exception(e.getMessage() + " " + item.toString());
				} catch (Exception e) {
					throw new Exception(e.getMessage());
				}
			}

		// if no item was found
		if (item == null)
			throw new Exception("No code to be executed: " + current.toString());

		return Status.OK;
	}

	private static boolean isStatement(Item i) {
		if (!(i instanceof Text))
			return false;

		Frame frame = i.getParent();

		if (i == frame.getTitle() || i == frame.getName() || i.isAnnotation()) {
			return false;
		}

		return true;
	}

	/**
	 * This method should be modified to parse strings correctly
	 * 
	 * @param statement
	 * @return
	 */
	private static String[] parseStatement(String statement) {
		return statement.split(TOKEN_SEPARATOR);
	}

	/**
	 * Runs a SIMPLE procedure statement.
	 * 
	 * @param tokens
	 *            the parsed Call statement.
	 * @param code
	 *            the expeditee item containing the procedure call and the link
	 *            to the frame with the procedure code.
	 * @param context
	 *            the current context from which the procedure call is being
	 *            made.
	 * @throws Exception
	 *             when errors occur in running the procedure.
	 */
	private static Status Call(String[] tokens, Text code, Context context)
			throws Exception {
		// Check that the call statement is linked
		if (code.getLink() == null)
			throw new Exception("Unlinked call statement: " + code.toString());

		Frame procedure = FrameIO.LoadFrame(code.getAbsoluteLink());

		// Add call to the start of the title if it doesnt exist
		// This makes the call and signature tokens counts match
		String procedureTitle = procedure.getTitle().getFirstLine();
		if (!procedureTitle.toLowerCase().startsWith("call "))
			procedureTitle = "call " + procedureTitle;

		// Check that the calling statement matches the procedure
		// signature
		String[] procedureSignature = procedureTitle.split(TOKEN_SEPARATOR);

		// Check for the right amount of parametres
		if (procedureSignature.length < tokens.length)
			throw new Exception("Call statement has too many parametres: "
					+ code.toString());
		else if (procedureSignature.length > tokens.length)
			throw new Exception("Call statement has too few parametres: "
					+ code.toString());
		// else if (procedureSignature[1].equals(tokens[1]))
		// throw new Exception("Call statement and procedure name dont match: "
		// + code.toString());

		// create the new context for the sub procedure
		Context newContext = new Context();
		// Check that the types/prefixes match
		for (int i = 2; i < tokens.length; i++) {
			// TODO allow for auto casting of primitives
			if (tokens[i].substring(1, 2).equals(
					procedureSignature[i].substring(1, 2))) {
				// Add the variables to the new context
				if (Primitives.isPrimitive(tokens[i])) {
					try {
						// try and get the value for the variable
						SPrimitive p = context.getPrimitives().getVariable(
								tokens[i]);
						newContext.getPrimitives()
								.add(procedureSignature[i], p);
					} catch (Exception e) {
						// If an exception occurs the variable doesnt
						// exist in the current context
						// So the variable must be added to both context
						context.getPrimitives().add(tokens[i], new SString(""));
						newContext.getPrimitives().add(procedureSignature[i],
								new SString(""));
					}
				} else if (Pointers.isPointer(tokens[i])) {
					try {
						// try and get the value for the variable
						SPointer p = context.getPointers().getVariable(
								tokens[i]);
						newContext.getPointers().add(procedureSignature[i], p);
					} catch (Exception e) {
						// If an exception occurs the variable doesnt
						// exist in the current context
						// So the variable must be added to both context
						context.getPointers().setObject(tokens[i], null);
						newContext.getPointers().setObject(
								procedureSignature[i], null);
					}
				} else
					throw new Exception("Unrecognised variable type: "
							+ tokens[i] + " in " + code.toString());
			} else
				throw new IncorrectTypeException(procedureSignature[i], i);
		}

		// Follow the link and Run the code for the procedure
		Status result = RunFrameAndReportError(code, newContext);
		// If a return statement ends the procedure then we accept this as
		// normal execution
		switch (result) {
		case Return:
			result = Status.OK;
			break;
		case Break:
			throw new Exception(BREAK_TEXT + " statement without matching "
					+ LOOP_TEXT + " in " + code.toString());
		case Continue:
			throw new Exception("");
		}

		// Now copy the values from the procedure context into the
		// current context
		for (int i = 2; i < tokens.length; i++) {
			try {
				if (Primitives.isPrimitive(tokens[i])) {
					// try and get the value for the variable
					SVariable p = context.getPrimitives()
							.getVariable(tokens[i]);
					SVariable newP = newContext.getPrimitives().getVariable(
							procedureSignature[i]);
					p.setValue(newP);
				} else {
					// try and get the value for the variable
					SVariable p = context.getPointers().getVariable(tokens[i]);
					SVariable newP = newContext.getPointers().getVariable(
							procedureSignature[i]);
					p.setValue(newP);
				}
			} catch (Exception e) {
				assert (false);
			}
		}

		return result;
	}

	/**
	 * Runs a text item on a frame as a SIMPLE statement. The statement is
	 * parsed and if it is a recognised SIMPLE keyword or procedure the code is
	 * executed.
	 * 
	 * @param code
	 *            the item containing the code to be executed.
	 * @param context
	 * @return
	 * @throws Exception
	 */
	private static Status RunItem(Text code, Context context,
			Status lastItemStatus) throws Exception {
		String codeText = code.getTextNoList();
		// Comments without links are a no-op
		if (codeText.startsWith("#") || codeText.startsWith("//")) {
			if (code.getLink() == null)
				return Status.OK;
			else
				return RunFrameAndReportError(code, context);
		}

		// Now check for code statements
		codeText = codeText.trim();
		String[] tokens = parseStatement(codeText.toLowerCase());
		// At present only set statements can have literals so they
		// are the only statements that we have to worry about string
		// literals
		assert (tokens.length > 0);
		if (tokens[0].equals("call") && tokens.length >= 2) {
			return Call(tokens, code, context);
			// Check if the user wants to display a message
			// Check for set statements
		} else if (tokens[0].startsWith("set")) {
			if (tokens[0].equals("set")) {
				try {
					// Check if we are setting variable to variable
					if (tokens[2].startsWith(SVariable.prefix)
							&& tokens.length == 3) {
						context.getPrimitives().set(tokens[1], tokens[2]);
						return Status.OK;
					}
					// Otherwise we are setting a variable with a literal
					else {
						// check for strings enclosed in quotes
						if (tokens[2].startsWith("\"")) {
							if (tokens[tokens.length - 1].endsWith("\"")) {
								context.getPrimitives().setValue(
										tokens[1],
										new SString(codeText.substring(codeText
												.indexOf("\"") + 1, codeText
												.length() - 1)));
								return Status.OK;
							} else
								throw new Exception("Expected matching \" "
										+ code.toString());
							// set a literal
						} else if (tokens.length == 3) {
							context.getPrimitives().setValue(tokens[1],
									tokens[2]);
							return Status.OK;
						}
					}
				} catch (Exception e) {
					throw new RuntimeException(e.getMessage());
				}
			} else if (tokens[0].startsWith("setitem")) {
				if (tokens[0].equals("setitemposition")) {
					assertVariableType(tokens[1], 1, SPointer.itemPrefix);
					// assertPrimitiveType(tokens[2], 2);
					// assertPrimitiveType(tokens[3], 3);
					Item item = (Item) context.getPointers().getVariable(
							tokens[1]).getValue();
					item.setPosition(context.getPrimitives().getVariable(
							tokens[2]).integerValue().intValue(), context
							.getPrimitives().getVariable(tokens[3])
							.integerValue().intValue());
				} else if (tokens[0].equals("setitemthickness")) {
					assertVariableType(tokens[1], 1, SPointer.itemPrefix);
					// assertPrimitiveType(tokens[2], 2);
					Item item = (Item) context.getPointers().getVariable(
							tokens[1]).getValue();
					item.setThickness(context.getPrimitives().getVariable(
							tokens[2]).integerValue().intValue());
				} else if (tokens[0].equals("setitemwidth")) {
					assertVariableType(tokens[1], 1, SPointer.itemPrefix);
					// assertPrimitiveType(tokens[2], 2);
					Item item = (Item) context.getPointers().getVariable(
							tokens[1]).getValue();
					item.setWidth(context.getPrimitives()
							.getVariable(tokens[2]).integerValue().intValue());
				} else if (tokens[0].equals("setitemsize")) {
					assertVariableType(tokens[1], 1, SPointer.itemPrefix);
					// assertPrimitiveType(tokens[2], 2);
					Item item = (Item) context.getPointers().getVariable(
							tokens[1]).getValue();
					item.setSize(context.getPrimitives().getVariable(tokens[2])
							.integerValue().intValue());
				} else if (tokens[0].equals("setitemlink")) {
					assertVariableType(tokens[1], 1, SPointer.itemPrefix);
					// assertPrimitiveType(tokens[2], 2);
					Item item = (Item) context.getPointers().getVariable(
							tokens[1]).getValue();
					item.setLink(context.getPrimitives().getVariable(tokens[2])
							.stringValue());
				} else if (tokens[0].equals("setitemfillcolor")) {
					assertVariableType(tokens[1], 1, SPointer.itemPrefix);
					// assertPrimitiveType(tokens[2], 2);
					Item item = (Item) context.getPointers().getVariable(
							tokens[1]).getValue();
					String stringColor = context.getPrimitives().getVariable(
							tokens[2]).stringValue();
					item.setBackgroundColor((Color) Conversion.Convert(
							Color.class, stringColor));
				} else if (tokens[0].equals("setitemcolor")) {
					assertVariableType(tokens[1], 1, SPointer.itemPrefix);
					// assertPrimitiveType(tokens[2], 2);
					Item item = (Item) context.getPointers().getVariable(
							tokens[1]).getValue();
					String stringColor = context.getPrimitives().getVariable(
							tokens[2]).stringValue();
					item.setColor((Color) Conversion.Convert(Color.class,
							stringColor));
				} else if (tokens[0].equals("setitemtext")) {
					assertVariableType(tokens[1], 1, SPointer.itemPrefix);
					// assertPrimitiveType(tokens[2], 2);
					String newText = context.getPrimitives().getVariable(
							tokens[2]).stringValue();
					Text textItem = (Text) context.getPointers().getVariable(
							tokens[1]).getValue();
					textItem.setText(newText);
				} else
					throw new Exception("Unsupported setItem command: "
							+ code.toString());
				return Status.OK;
			} else if (tokens[0].equals("setstrchar")) {
				assertExactParametreCount(tokens, 3);
				StringBuffer s = new StringBuffer(context.getPrimitives()
						.getStringValue(tokens[1]));
				int pos = (int) context.getPrimitives().getIntegerValue(
						tokens[2]);
				char newChar = context.getPrimitives().getCharacterValue(
						tokens[3]);
				while (pos > s.length()) {
					s.append(newChar);
				}
				s.setCharAt(pos - 1, newChar);

				context.getPrimitives().setValue(tokens[1],
						new SString(s.toString()));
				return Status.OK;
			} else if (tokens[0].equals("setcharinitem")) {
				assertExactParametreCount(tokens, 4);
				assertVariableType(tokens[1], 1, SPointer.itemPrefix);

				int row = (int) context.getPrimitives().getIntegerValue(
						tokens[3]) - 1;
				int col = (int) context.getPrimitives().getIntegerValue(
						tokens[4]) - 1;
				char newChar = context.getPrimitives().getCharacterValue(
						tokens[2]);
				Text item = (Text) context.getPointers().getVariable(tokens[1])
						.getValue();
				List<String> itemText = item.getText();

				while (row >= itemText.size()) {
					StringBuffer sb = new StringBuffer();
					for (int i = 0; i <= col; i++)
						sb.append(newChar);
					itemText.add(sb.toString());
				}
				StringBuffer sb = new StringBuffer(itemText.get(row));
				while (sb.length() <= col)
					sb.append(newChar);
				sb.setCharAt(col, newChar);

				// set the modified string
				itemText.set(row, sb.toString());
				item.setTextList(itemText);

				return Status.OK;
			}
		} else if (tokens[0].startsWith("assert")) {
			if (tokens[0].equals("asserttrue")) {
				assertExactParametreCount(tokens, 1);
				if (!context.getPrimitives().getBooleanValue(tokens[1]))
					throw new UnitTestFailedException("true", "false");
				return Status.OK;
			} else if (tokens[0].equals("assertfalse")) {
				assertExactParametreCount(tokens, 1);
				if (context.getPrimitives().getBooleanValue(tokens[1]))
					throw new UnitTestFailedException("false", "true");
				return Status.OK;
			} else if (tokens[0].equals("assertnull")) {
				assertExactParametreCount(tokens, 1);
				Object value = context.getPrimitives().getVariable(tokens[1])
						.getValue();
				if (value != null)
					throw new UnitTestFailedException("null", value.toString());
				return Status.OK;
			} else if (tokens[0].equals("assertequals")) {
				assertExactParametreCount(tokens, 2);
				if (!context.getPrimitives().equalValues(tokens[1], tokens[2]))
					throw new UnitTestFailedException(context.getPrimitives()
							.getStringValue(tokens[1]), context.getPrimitives()
							.getStringValue(tokens[2]));
				return Status.OK;
			} else if (tokens[0].equals("assertdefined")) {
				assertExactParametreCount(tokens, 1);
				if (!context.isDefined(tokens[1])) {
					throw new UnitTestFailedException(tokens[1] + " exists",
							"not defined");
				}
				return Status.OK;
			}
		} else if (tokens[0].startsWith("goto")) {
			String frameNameVar = DEFAULT_STRING;
			if (tokens.length > 1) {
				assertExactParametreCount(tokens, 1);
				frameNameVar = tokens[1];
			}
			String frameName = context.getPrimitives().getStringValue(
					frameNameVar);
			NavigationActions.Goto(frameName);
			return Status.OK;
		} else if (tokens[0].startsWith("get")) {
			if (tokens[0].startsWith("getframe")) {
				if (tokens[0].equals("getframevalue")) {
					assertMinParametreCount(tokens, 4);
					assertVariableType(tokens[1], 1, SPointer.framePrefix);

					// Get the attribute to be searched for on the target frame
					Frame targetFrame = (Frame) context.getPointers()
							.getVariable(tokens[1]).getValue();
					String targetAttribute = context.getPrimitives()
							.getStringValue(tokens[2]).toLowerCase()
							+ ":";
					Boolean found = false;
					String value = "";
					Text attributeItem = null;
					Text valueItem = null;
					// Begin the search
					for (Text text : targetFrame.getBodyTextItems()) {
						String s = text.getTextNoList().toLowerCase();
						if (s.startsWith(targetAttribute)) {
							attributeItem = text;
							value = AttributeUtils.stripValue(s);
							if (value.length() > 0) {
								found = true;
							}
							break;
						}
					}
					// Keep looking for a matching value nearby if we found an
					// attribute without the value in the same item
					if (!found && attributeItem != null) {
						Point endPoint = attributeItem
								.getEndParagraphPosition();

						for (Text text : targetFrame.getBodyTextItems()) {
							Point startPoint = text.getPosition();
							if (Math.abs(startPoint.y - endPoint.y) < 10
									&& Math.abs(startPoint.x - endPoint.x) < 20) {
								found = true;
								valueItem = text;
								value = text.getTextNoList();
								break;
							}
						}
					}

					// Set the values of the output parametres
					context.getPrimitives().setValue(tokens[3],
							new SString(value));
					context.getPrimitives().setValue(tokens[4],
							new SBoolean(found));
					if (tokens.length > 5) {
						context.getPointers().setObject(tokens[5],
								attributeItem);
						if (tokens.length > 6) {
							context.getPointers().setObject(tokens[6],
									valueItem);
						}
					}

					return Status.OK;
				} else if (tokens[0].startsWith("getframename")) {
					String frameNameVar = DEFAULT_STRING;
					String frameVar = DEFAULT_FRAME;

					if (tokens.length > 1) {
						assertExactParametreCount(tokens, 2);
						assertVariableType(tokens[1], 1, SPointer.framePrefix);
						frameNameVar = tokens[2];
						frameVar = tokens[1];
					}
					Frame frame = (Frame) context.getPointers().getVariable(
							frameVar).getValue();
					context.getPrimitives().setValue(frameNameVar,
							frame.getFrameName());
					return Status.OK;
				}
			} else if (tokens[0].startsWith("getcurrent")) {
				if (tokens[0].equals("getcurrentframe")) {
					assertMinParametreCount(tokens, 1);
					assertVariableType(tokens[1], 1, SPointer.framePrefix);

					Frame currentFrame = DisplayIO.getCurrentFrame();
					context.getPointers().setObject(tokens[1], currentFrame);

					// check if the user is also after the frameName
					if (tokens.length > 2) {
						context.getPrimitives().setValue(tokens[2],
								new SString(currentFrame.getFrameName()));
					}
					return Status.OK;
				}
			} else if (tokens[0].startsWith("getitem")) {
				if (tokens[0].equals("getitemposition")) {
					assertVariableType(tokens[1], 1, SPointer.itemPrefix);
					// assertPrimitiveType(tokens[2], 2);
					// assertPrimitiveType(tokens[3], 3);
					Point pos = ((Item) context.getPointers().getVariable(
							tokens[1]).getValue()).getPosition();
					Integer x = pos.x;
					Integer y = pos.y;
					context.getPrimitives()
							.setValue(tokens[2], new SInteger(x));
					context.getPrimitives()
							.setValue(tokens[3], new SInteger(y));
				} else if (tokens[0].equals("getitemthickness")) {
					assertVariableType(tokens[1], 1, SPointer.itemPrefix);
					// assertPrimitiveType(tokens[2], 2);
					Float thickness = ((Item) context.getPointers()
							.getVariable(tokens[1]).getValue()).getThickness();
					context.getPrimitives().setValue(tokens[2],
							new SReal(thickness));
				} else if (tokens[0].equals("getitemwidth")) {
					assertVariableType(tokens[1], 1, SPointer.itemPrefix);
					// assertPrimitiveType(tokens[2], 2);
					Integer width = ((Item) context.getPointers().getVariable(
							tokens[1]).getValue()).getWidth();
					context.getPrimitives().setValue(tokens[2],
							new SInteger(width));
				} else if (tokens[0].equals("getitemsize")) {
					assertVariableType(tokens[1], 1, SPointer.itemPrefix);
					// assertPrimitiveType(tokens[2], 2);
					Integer size = ((Item) context.getPointers().getVariable(
							tokens[1]).getValue()).getSize();
					context.getPrimitives().setValue(tokens[2],
							new SInteger(size));
				} else if (tokens[0].equals("getitemlink")) {
					assertVariableType(tokens[1], 1, SPointer.itemPrefix);
					// assertPrimitiveType(tokens[2], 2);
					String link = ((Item) context.getPointers().getVariable(
							tokens[1]).getValue()).getLink();
					context.getPrimitives().setValue(tokens[2],
							new SString(link));
				} else if (tokens[0].equals("getitemfillcolor")) {
					assertVariableType(tokens[1], 1, SPointer.itemPrefix);
					// assertPrimitiveType(tokens[2], 2);
					Color itemColor = ((Item) context.getPointers()
							.getVariable(tokens[1]).getValue())
							.getPaintBackgroundColor();
					String color = itemColor.getRed() + " "
							+ itemColor.getGreen() + " " + itemColor.getBlue();
					context.getPrimitives().setValue(tokens[2],
							new SString(color));
				} else if (tokens[0].equals("getitemcolor")) {
					assertVariableType(tokens[1], 1, SPointer.itemPrefix);
					// assertPrimitiveType(tokens[2], 2);
					Color itemColor = ((Item) context.getPointers()
							.getVariable(tokens[1]).getValue()).getPaintColor();
					String color = itemColor.getRed() + " "
							+ itemColor.getGreen() + " " + itemColor.getBlue();
					context.getPrimitives().setValue(tokens[2],
							new SString(color));
				} else if (tokens[0].equals("getitemtext")) {
					assertVariableType(tokens[1], 1, SPointer.itemPrefix);
					Item item = ((Item) context.getPointers().getVariable(
							tokens[1]).getValue());
					context.getPrimitives().setValue(
							tokens[2],
							new SString((item instanceof Text) ? ((Text) item)
									.getTextNoList() : ""));
				} else
					throw new Exception("Unsupported getItem command: "
							+ code.toString());

				return Status.OK;
			} else if (tokens[0].equals("getstrchar")) {
				assertExactParametreCount(tokens, 3);
				String s = context.getPrimitives().getStringValue(tokens[1]);
				int pos = (int) context.getPrimitives().getIntegerValue(
						tokens[2]);

				context.getPrimitives().setValue(tokens[3],
						new SCharacter(s.charAt(pos - 1)));
				return Status.OK;
			} else if (tokens[0].equals("getstrlength")) {
				assertExactParametreCount(tokens, 2);
				String s = context.getPrimitives().getStringValue(tokens[1]);

				context.getPrimitives().setValue(tokens[2],
						new SInteger(s.length()));
				return Status.OK;
			} else if (tokens[0].equals("getelapsedtime")) {
				assertExactParametreCount(tokens, 1);
				context.getPrimitives().setValue(tokens[1],
						new SInteger(AgentStats.getMilliSecondsElapsed()));
				return Status.OK;
			}
		} else if (tokens[0].equals("or")) {
			for (int i = 1; i < tokens.length - 1; i++) {
				if (Primitives.isPrimitive(tokens[i])) {
					if (context.getPrimitives().getBooleanValue(tokens[i])) {
						context.getPrimitives().setValue(
								tokens[tokens.length - 1], new SBoolean(true));
						return Status.OK;
					}
				}
			}
			context.getPrimitives().setValue(tokens[tokens.length - 1],
					new SBoolean(false));
			return Status.OK;
		} else if (tokens[0].equals("and")) {
			for (int i = 1; i < tokens.length - 1; i++) {
				if (Primitives.isPrimitive(tokens[i])) {
					if (!context.getPrimitives().getBooleanValue(tokens[i])) {
						context.getPrimitives().setValue(
								tokens[tokens.length - 1], new SBoolean(false));
						return Status.OK;
					}
				}
			}
			context.getPrimitives().setValue(tokens[tokens.length - 1],
					new SBoolean(true));
			return Status.OK;
		} else if (tokens[0].equals("messagelnitem")
				|| tokens[0].equals("messagelineitem")) {
			String itemVar = DEFAULT_ITEM;

			if (tokens.length > 1) {
				assertExactParametreCount(tokens, 1);
				itemVar = tokens[1];
				assertVariableType(itemVar, 1, SPointer.itemPrefix);
			}
			Item message = (Item) context.getPointers().getVariable(itemVar)
					.getValue();
			try {
				FrameGraphics.DisplayMessage(((Text) message).copy());
			} catch (NullPointerException e) {
				// Just ignore not text items!
				FrameGraphics.DisplayMessage("null");
			} catch (ClassCastException e) {
				// Just ignore not text items!
				FrameGraphics.DisplayMessage(message.toString());
			} catch (Exception e) {
				// Just ignore other errors
			}
			return Status.OK;
		} else if (tokens[0].equals("messageln")
				|| tokens[0].equals("messageline")
				|| tokens[0].equals("errorln") || tokens[0].equals("errorline")) {
			StringBuilder message = new StringBuilder();
			if (tokens.length == 1) {
				message.append(context.getPrimitives().getVariable(
						DEFAULT_STRING).stringValue()
						+ " ");
			} else {
				for (int i = 1; i < tokens.length; i++) {
					if (Primitives.isPrimitive(tokens[i])) {
						message.append(context.getPrimitives().getVariable(
								tokens[i]).stringValue()
								+ " ");
					} else
						throw new Exception("Illegal parametre: [" + tokens[i]
								+ "] in line " + code.toString());
				}
			}

			if (tokens[0].equals("errorln") || tokens[0].equals("errorline"))
				FrameGraphics.ErrorMessage(message.toString());
			else
				FrameGraphics.DisplayMessageAlways(message.toString());
			return Status.OK;
		} else if (tokens[0].startsWith("else")) {
			// if the if statement was false then run the else statement
			if (lastItemStatus == Status.FalseIf) {
				// check if it is a one line if statment
				if (tokens.length > 1) {
					// put together the one line statement
					StringBuilder statement = new StringBuilder();
					for (int i = 1; i < tokens.length; i++)
						statement.append(tokens[i]).append(' ');
					// create a copy of the code item to run
					Text copiedCode = code.copy();
					copiedCode.setText(statement.toString());
					return RunItem(copiedCode, context, Status.OK);
				} else {
					return RunFrameAndReportError(code, context);
				}
			} else if (lastItemStatus == Status.TrueIf) {
				return Status.OK;
			}
			throw new RuntimeException("Else without matching If statement");
		} else if (tokens[0].startsWith("if")) {
			Boolean result = null;
			int parametres = 1;
			String variable = DEFAULT_ITEM;
			String ifStatement = tokens[0];
			// Set the default variable
			if (tokens.length == 1) {
				if (ifStatement.equals("if") || ifStatement.equals("ifnot"))
					variable = DEFAULT_STRING;
				else if (ifStatement.equals("ifzero")
						|| ifStatement.equals("ifnotzero"))
					variable = DEFAULT_INTEGER;
			} else {
				variable = tokens[1];
			}

			if (ifStatement.equals("if")) {
				result = context.getPrimitives().getBooleanValue(variable);
			} else if (ifStatement.equals("ifnot")) {
				result = !context.getPrimitives().getBooleanValue(variable);
			} else if (ifStatement.equals("ifdefined")) {
				result = context.isDefined(tokens[1]);
			} else if (ifStatement.equals("ifnotdef")) {
				result = !context.isDefined(tokens[1]);
			} else if (ifStatement.equals("ifzero")) {
				result = context.getPrimitives().getIntegerValue(variable) == 0;
			} else if (ifStatement.equals("ifnotzero")) {
				result = context.getPrimitives().getIntegerValue(variable) != 0;
			} else if (tokens[0].equals("ifeq")) {
				result = context.equalValues(tokens[1], tokens[2]);
				parametres = 2;
			} else if (tokens[0].equals("ifeqnocase")) {
				result = context.getPrimitives().equalValuesNoCase(tokens[1],
						tokens[2]);
				parametres = 2;
			} else if (tokens[0].equals("ifnoteqnocase")) {
				result = !context.getPrimitives().equalValuesNoCase(tokens[1],
						tokens[2]);
				parametres = 2;
			} else if (tokens[0].equals("ifnoteq")) {
				result = !context.equalValues(tokens[1], tokens[2]);
				parametres = 2;
			} else if (tokens[0].equals("ifless")) {
				result = context.getPrimitives().compareValues(tokens[1],
						tokens[2]) < 0;
				parametres = 2;
			} else if (tokens[0].equals("ifgtr")) {
				result = context.getPrimitives().compareValues(tokens[1],
						tokens[2]) > 0;
				parametres = 2;
			} else if (tokens[0].equals("ifgeq")) {
				result = context.getPrimitives().compareValues(tokens[1],
						tokens[2]) >= 0;
				parametres = 2;
			} else if (tokens[0].equals("ifleq")) {
				result = context.getPrimitives().compareValues(tokens[1],
						tokens[2]) <= 0;
				parametres = 2;
			} else {
				// assertVariableType(variable, 1, SPointer.itemPrefix);
				if (ifStatement.equals("ifannotation")) {
					result = ((Item) context.getPointers()
							.getVariable(variable).getValue()).isAnnotation();
				} else if (ifStatement.equals("iflinked")) {
					result = ((Item) context.getPointers()
							.getVariable(variable).getValue()).getLink() != null;
				} else if (ifStatement.equals("ifactioned")) {
					result = ((Item) context.getPointers()
							.getVariable(variable).getValue()).getAction() != null;
				} else if (ifStatement.equals("ifnotactioned")) {
					result = ((Item) context.getPointers()
							.getVariable(variable).getValue()).getAction() == null;
				} else if (ifStatement.equals("ifbodytext")) {
					Item i = (Item) context.getPointers().getVariable(variable)
							.getValue();
					result = i instanceof Text && !i.isFrameName()
							&& !i.isFrameTitle();
				} else if (ifStatement.equals("ifnotannotation")) {
					result = !((Item) context.getPointers().getVariable(
							variable).getValue()).isAnnotation();
				} else if (ifStatement.equals("ifnotlinked")) {
					result = ((Item) context.getPointers()
							.getVariable(variable).getValue()).getLink() == null;
				} else if (ifStatement.equals("ifnotbodytext")) {
					Item i = (Item) context.getPointers().getVariable(variable)
							.getValue();
					result = !(i instanceof Text) || i.isFrameName()
							|| i.isFrameTitle();
				}
			}

			if (result == null)
				throw new RuntimeException("Invalid If statement");
			// Now check if we need to run the code
			else if (result) {
				Status status;
				// check if it is a one line if statment
				if (tokens.length > parametres + 1) {
					// put together the one line statement
					StringBuilder statement = new StringBuilder();
					for (int i = parametres + 1; i < tokens.length; i++)
						statement.append(tokens[i]).append(' ');
					// create a copy of the code item to run
					Text copiedCode = code.copy();
					copiedCode.setText(statement.toString());
					status = RunItem(copiedCode, context, Status.OK);
				} else {
					status = RunFrameAndReportError(code, context);
				}
				if (status == Status.OK) {
					return Status.TrueIf;
				} else {
					return status;
				}
			}
			return Status.FalseIf;

		} // Look for variable length methods
		else if (tokens[0].equals("attachitemtocursor")) {
			String itemVar = DEFAULT_ITEM;

			if (tokens.length > 1) {
				assertExactParametreCount(tokens, 1);
				assertVariableType(tokens[1], 1, SPointer.itemPrefix);
				itemVar = tokens[1];
			}
			Item item = (Item) context.getPointers().getVariable(itemVar)
					.getValue();
			FrameMouseActions.pickup(item);
			return Status.OK;
		} else if (tokens[0].equals("attachstrtocursor")) {
			String stringVar = DEFAULT_STRING;

			if (tokens.length > 1) {
				assertExactParametreCount(tokens, 1);
				stringVar = tokens[1];
			}
			String s = context.getPrimitives().getStringValue(stringVar);
			Frame frame = DisplayIO.getCurrentFrame();
			Text item = frame.createNewText(s);
			FrameMouseActions.pickup(item);
			// DisplayIO.repaint();
			return Status.OK;
		} else if (tokens[0].equals("additemtoframe")) {
			String itemVar = DEFAULT_ITEM;
			String frameVar = DEFAULT_FRAME;

			if (tokens.length > 1) {
				assertExactParametreCount(tokens, 2);
				assertVariableType(tokens[1], 1, SPointer.framePrefix);
				assertVariableType(tokens[2], 2, SPointer.itemPrefix);
				itemVar = tokens[2];
				frameVar = tokens[1];
			}
			Frame frame = (Frame) context.getPointers().getVariable(frameVar)
					.getValue();
			Item item = (Item) context.getPointers().getVariable(itemVar)
					.getValue();
			frame.addItem(item);

			return Status.OK;
		} else if (tokens[0].equals("connectdots")) {
			assertMinParametreCount(tokens, 2);
			assertVariableType(tokens[1], 1, SPointer.itemPrefix);
			assertVariableType(tokens[2], 2, SPointer.itemPrefix);
			Dot dot1 = null;
			Dot dot2 = null;
			try {
				dot1 = (Dot) context.getPointers().getVariable(tokens[1])
						.getValue();
			} catch (Exception e) {
				throw new IncorrectTypeException("Dot", 1);
			}
			try {
				dot2 = (Dot) context.getPointers().getVariable(tokens[2])
						.getValue();
			} catch (Exception e) {
				throw new IncorrectTypeException("Dot", 2);
			}
			Frame frame = dot1.getParent();
			frame.addItem(new Line(dot1, dot2, frame.getNextItemID()));
			return Status.OK;
		} else if (tokens[0].equals("createitem")) {
			assertMinParametreCount(tokens, 4);
			assertVariableType(tokens[1], 1, SPointer.framePrefix);
			assertVariableType(tokens[4], 4, SPointer.itemPrefix);
			Frame frame = (Frame) context.getPointers().getVariable(tokens[1])
					.getValue();
			Item newItem;
			int x = (int) context.getPrimitives().getIntegerValue(tokens[2]);
			int y = (int) context.getPrimitives().getIntegerValue(tokens[3]);
			// check for the option text and action for the new item
			if (tokens.length > 5) {
				String newText = context.getPrimitives().getStringValue(
						tokens[5]);
				String newAction = null;
				if (tokens.length > 6) {
					newAction = context.getPrimitives().getStringValue(
							tokens[6]);
				}
				newItem = frame.addText(x, y, newText, newAction);
			} else
				// create a point if the optional params are not provided
				newItem = frame.addDot(x, y);
			context.getPointers().setObject(tokens[4], newItem);

			return Status.OK;
		} else if (tokens[0].equals("createframe")) {

			String framesetName = DEFAULT_STRING;
			String frameVar = DEFAULT_FRAME;

			if (tokens.length > 1) {
				assertMinParametreCount(tokens, 2);
				assertVariableType(tokens[2], 2, SPointer.framePrefix);
				framesetName = tokens[1];
				frameVar = tokens[2];
			}

			if (tokens.length > 3) {
				context.createFrame(framesetName, frameVar, tokens[3]);
			} else
				context.createFrame(framesetName, frameVar, null);

			return Status.OK;
		} else if (tokens[0].equals("closeframe")) {
			String frameVar = DEFAULT_FRAME;

			if (tokens.length > 1) {
				assertVariableType(tokens[1], 1, SPointer.framePrefix);
				frameVar = tokens[1];
			}

			if (tokens.length > 2) {
				// assertPrimitiveType(tokens[3], 3);
				context.closeFrame(frameVar, tokens[3]);
			} else
				context.closeFrame(frameVar, null);

			return Status.OK;
		} else if (tokens[0].equals("readframe")
				|| tokens[0].equals("openframe")) {

			String frameName = DEFAULT_STRING;
			String frameVar = DEFAULT_FRAME;

			if (tokens.length > 1) {
				assertMinParametreCount(tokens, 2);
				assertVariableType(tokens[2], 2, SPointer.framePrefix);
				frameName = tokens[1];
				frameVar = tokens[2];
				// assertPrimitiveType(frameName, 1);
			}

			if (tokens.length > 3) {
				// assertPrimitiveType(tokens[3], 3);
				context.readFrame(frameName, frameVar, tokens[3]);
			} else
				context.readFrame(frameName, frameVar, null);

			return Status.OK;
		} else if (tokens[0].equals("readkbdcond")) {

			String nextCharVarName = DEFAULT_CHAR;
			String wasCharVarName = DEFAULT_BOOLEAN;

			if (tokens.length > 1) {
				assertMinParametreCount(tokens, 2);
				assertVariableType(tokens[2], 2, SPointer.framePrefix);
				nextCharVarName = tokens[1];
				wasCharVarName = tokens[2];
			}

			Character nextChar = _KeyStrokes.poll();
			boolean hasChar = nextChar != null;
			context.getPrimitives().setValue(wasCharVarName,
					new SBoolean(hasChar));
			if (hasChar)
				context.getPrimitives().setValue(nextCharVarName,
						new SCharacter(nextChar));

			return Status.OK;
		} else if (tokens[0].equals("openwritefile")) {
			assertVariableType(tokens[1], 1, SString.prefix);
			assertVariableType(tokens[2], 2, SPointer.filePrefix);

			if (tokens.length > 3) {
				assertVariableType(tokens[3], 3, SBoolean.prefix);
				context.openWriteFile(tokens[1], tokens[2], tokens[3]);
			} else
				context.openWriteFile(tokens[1], tokens[2]);

			return Status.OK;
		} else if (tokens[0].equals("writefile")
				|| tokens[0].equals("writelinefile")
				|| tokens[0].equals("writelnfile")) {
			assertVariableType(tokens[1], 1, SPointer.filePrefix);

			StringBuilder textToWrite = new StringBuilder();
			if (tokens.length == 1) {
				textToWrite.append(context.getPrimitives().getVariable(
						DEFAULT_STRING).stringValue()
						+ " ");
			} else {
				for (int i = 2; i < tokens.length; i++) {
					if (Primitives.isPrimitive(tokens[i])) {
						textToWrite.append(context.getPrimitives().getVariable(
								tokens[i]).stringValue());
					} else
						throw new Exception("Illegal parametre: " + tokens[i]
								+ " in " + code.toString());
				}
			}

			if (!tokens[0].equals("writefile"))
				textToWrite.append('\n');
			context.writeFile(tokens[1], textToWrite.toString());

			return Status.OK;
		} else if (tokens[0].equals("concatstr")) {
			assertMinParametreCount(tokens, 3);
			String resultVar = tokens[tokens.length - 1];

			StringBuilder sb = new StringBuilder();
			// loop through all the strings concatenating them
			for (int i = 1; i < tokens.length - 1; i++) {
				// assertPrimitiveType(tokens[i], i);
				sb.append(context.getPrimitives().getStringValue(tokens[i]));
			}
			context.getPrimitives().setValue(resultVar,
					new SString(sb.toString()));
			return Status.OK;
		} else if (tokens[0].equals("convstrlower")) {
			assertExactParametreCount(tokens, 1);
			// assertPrimitiveType(tokens[1], 1);
			context.getPrimitives().setValue(
					tokens[1],
					new SString(context.getPrimitives().getStringValue(
							tokens[1]).toLowerCase()));
			return Status.OK;
		} else if (tokens[0].equals("convstrupper")) {
			assertExactParametreCount(tokens, 1);
			// assertPrimitiveType(tokens[1], 1);
			context.getPrimitives().setValue(
					tokens[1],
					new SString(context.getPrimitives().getStringValue(
							tokens[1]).toUpperCase()));
			return Status.OK;
		} else if (tokens[0].equals("countcharsinstr")) {
			assertExactParametreCount(tokens, 3);
			String s = context.getPrimitives().getStringValue(tokens[1]);
			String pattern = context.getPrimitives().getStringValue(tokens[2]);
			int count = countCharsInString(s, pattern);
			context.getPrimitives().setValue(tokens[3], new SInteger(count));
			return Status.OK;
		} else if (tokens[0].equals("countcharsinitem")) {
			assertExactParametreCount(tokens, 3);
			assertVariableType(tokens[1], 1, SPointer.itemPrefix);
			Item item = (Item) context.getPointers().getVariable(tokens[1])
					.getValue();
			String pattern = context.getPrimitives().getStringValue(tokens[2]);
			int count = 0;
			if (item instanceof Text)
				count = countCharsInString(((Text) item).getTextNoList(),
						pattern);
			context.getPrimitives().setValue(tokens[3], new SInteger(count));
			return Status.OK;
		} else if (tokens[0].equals("parseframename")) {
			assertExactParametreCount(tokens, 4);
			String frameName = context.getPrimitives()
					.getStringValue(tokens[1]);
			String frameSet = "";
			int frameNo = -1;
			boolean success = true;
			try {
				frameSet = Conversion.getFrameset(frameName, false);
				frameNo = Conversion.getFrameNumber(frameName);
			} catch (Exception e) {
				success = false;
			}
			// assertPrimitiveType(tokens[2], 2);
			context.getPrimitives().setValue(tokens[2], new SBoolean(success));

			// assertPrimitiveType(tokens[3], 3);
			context.getPrimitives().setValue(tokens[3], new SString(frameSet));

			// assertPrimitiveType(tokens[4], 4);
			context.getPrimitives().setValue(tokens[4], new SInteger(frameNo));

			return Status.OK;
		} else if (tokens[0].equals("parsestr")) {
			assertMinParametreCount(tokens, 2);
			// assertPrimitiveType(tokens[1], 1);
			// assertPrimitiveType(tokens[2], 2);

			String s = context.getPrimitives().getStringValue(tokens[1]);

			String separator = context.getPrimitives()
					.getStringValue(tokens[2]);

			String[] split = s.split(separator, tokens.length - 4);

			if (tokens.length > 3) {
				// assertPrimitiveType(tokens[3], 3);
				int count = split.length;
				// if the string is not blank and its got a remainder then
				// decrease the count by 1 to account for the remainder
				if (split.length != 0 && split.length > tokens.length - 5)
					count--;

				context.getPrimitives()
						.setValue(tokens[3], new SInteger(count));

				if (tokens.length > 4) {
					// Set the remainder string
					// assertPrimitiveType(tokens[4], 4);
					if (split.length < tokens.length - 4)
						context.getPrimitives().setValue(tokens[4],
								new SString());
					else
						context.getPrimitives().setValue(tokens[4],
								new SString(split[split.length - 1]));

					// Set the strings for each of the vars
					if (tokens.length > 5) {
						for (int i = 5; i < tokens.length; i++) {
							// assertPrimitiveType(tokens[i], i);
							if (split.length < i - 4)
								context.getPrimitives().setValue(tokens[i],
										new SString());
							else
								context.getPrimitives().setValue(tokens[i],
										new SString(split[i - 5]));
						}
					}
				}
			}
			return Status.OK;
		} else if (tokens[0].equals("stripstr")) {
			assertExactParametreCount(tokens, 2);
			String s = context.getPrimitives().getStringValue(tokens[1]);
			String charsToStrip = context.getPrimitives().getStringValue(
					tokens[2]);
			for (int i = 0; i < charsToStrip.length(); i++)
				s = s.replaceAll(charsToStrip.substring(i, i + 1), "");
			context.getPrimitives().setValue(tokens[1], new SString(s));
			return Status.OK;
		} else if (tokens[0].equals("subststr")) {
			assertExactParametreCount(tokens, 3);
			String oldString = context.getPrimitives()
					.getStringValue(tokens[2]);
			String newString = context.getPrimitives()
					.getStringValue(tokens[3]);
			String result = context.getPrimitives().getStringValue(tokens[1]);
			result = result.replaceAll(oldString, newString);
			context.getPrimitives().setValue(tokens[1], new SString(result));
			return Status.OK;
		} else if (tokens[0].equals("substr")) {
			assertExactParametreCount(tokens, 4);
			int startPos = (int) context.getPrimitives().getIntegerValue(
					tokens[2]) - 1;
			int length = (int) context.getPrimitives().getIntegerValue(
					tokens[3]);
			String s = context.getPrimitives().getStringValue(tokens[1]);
			String result;
			if (startPos + length < s.length())
				result = s.substring(startPos, startPos + length);
			else
				result = s.substring(startPos);
			context.getPrimitives().setValue(tokens[4], new SString(result));
			return Status.OK;
		} else if (tokens[0].equals("pause")) {
			String lengthVar = DEFAULT_REAL;

			if (tokens.length > 1) {
				assertExactParametreCount(tokens, 1);
				lengthVar = tokens[1];
			}

			double time = context.getPrimitives().getDoubleValue(lengthVar);
			for (int i = 0; i < time * 10; i++) {
				Thread.yield();
				Thread.sleep(100);
			}
			return Status.OK;
		} else if (tokens[0].equals("glidecursorto")) {
			assertMinParametreCount(tokens, 2);
			int finalX = (int) context.getPrimitives().getIntegerValue(
					tokens[1]);
			int finalY = (int) context.getPrimitives().getIntegerValue(
					tokens[2]);
			int milliseconds = 1000;
			if (tokens.length > 3)
				milliseconds = (int) (context.getPrimitives().getDoubleValue(
						tokens[3]) * 1000);

			int initialX = DisplayIO.getMouseX();
			int initialY = DisplayIO.getMouseY();

			final int timeInterval = 40;

			int deltaX = (int) (finalX - initialX);
			int deltaY = (int) (finalY - initialY);

			int intervals = milliseconds / timeInterval;
			for (double i = 0; i < intervals; i++) {
				int newX = initialX + (int) (deltaX * i / intervals);
				int newY = initialY + (int) (deltaY * i / intervals);
				Thread.yield();
				Thread.sleep(timeInterval);
				DisplayIO.setCursorPosition(newX, newY);
				// DisplayIO.repaint();
			}
			// Thread.yield();
			Thread.sleep(milliseconds % timeInterval);
			DisplayIO.setCursorPosition(finalX, finalY);
			return Status.OK;
		}
		// Now look for fixed parametre statements
		else if (tokens.length == 1) {
			if (tokens[0].equals(EXIT_TEXT)) {
				return Status.Exit;
			} else if (tokens[0].equals(LOOP_TEXT)) {
				Status status = Status.OK;
				// Keep looping until break or exit occurs
				while (status == Status.OK || status == Status.Continue)
					status = RunFrameAndReportError(code, context);
				if (status == Status.Continue || status == Status.Break)
					status = Status.OK;
				return status;
			} else if (tokens[0].equals(CONTINUE_TEXT)) {
				return Status.Continue;
			} else if (tokens[0].equals(BREAK_TEXT)) {
				return Status.Break;
			} else if (tokens[0].equals(RETURN_TEXT)) {
				return Status.Return;
			} else if (tokens[0].equals("pressleftbutton")) {
				DisplayIO.pressMouse(InputEvent.BUTTON1_MASK);
				return Status.OK;
			} else if (tokens[0].equals("pressmiddlebutton")) {
				DisplayIO.pressMouse(InputEvent.BUTTON2_MASK);
				return Status.OK;
			} else if (tokens[0].equals("pressrightbutton")) {
				DisplayIO.pressMouse(InputEvent.BUTTON3_MASK);
				return Status.OK;
			} else if (tokens[0].equals("releaseleftbutton")) {
				DisplayIO.releaseMouse(InputEvent.BUTTON1_MASK);
				return Status.OK;
			} else if (tokens[0].equals("releasemiddlebutton")) {
				DisplayIO.releaseMouse(InputEvent.BUTTON2_MASK);
				return Status.OK;
			} else if (tokens[0].equals("releaserightbutton")) {
				DisplayIO.releaseMouse(InputEvent.BUTTON3_MASK);
				return Status.OK;
			} else if (tokens[0].equals("clickleftbutton")) {
				DisplayIO.clickMouse(InputEvent.BUTTON1_MASK);
				return Status.OK;
			} else if (tokens[0].equals("clickmiddlebutton")) {
				DisplayIO.clickMouse(InputEvent.BUTTON2_MASK);
				return Status.OK;
			} else if (tokens[0].equals("clickrightbutton")) {
				DisplayIO.clickMouse(InputEvent.BUTTON3_MASK);
				return Status.OK;
			} else if (tokens[0].equals("repaint")) {
				DisplayIO.repaint();
				return Status.OK;
			}

			if (tokens[0].equals("add")) {
				context.getPrimitives().add(DEFAULT_INTEGER);
			} else if (tokens[0].equals("subtract")) {
				context.getPrimitives().subtract(DEFAULT_INTEGER);
			} else
				throw new RuntimeException("Invalid statement:");
			return Status.OK;
		} else if (tokens.length == 2) {
			if (tokens[0].equals("closewritefile")) {
				assertVariableType(tokens[1], 1, SPointer.filePrefix);
				context.closeWriteFile(tokens[1]);

				return Status.OK;
			}

			if (Primitives.isPrimitive(tokens[1])) {
				if (tokens[0].equals("add")) {
					context.getPrimitives().add(tokens[1]);
				} else if (tokens[0].equals("subtract")) {
					context.getPrimitives().subtract(tokens[1]);
				} else
					throw new RuntimeException("Invalid statement:");
				return Status.OK;
			}
		} else if (tokens.length == 3) {
			if (tokens[0].equals("foreach")) {
				assertVariableType(tokens[1], 1, SPointer.itemPrefix);
				assertVariableType(tokens[2], 2, SPointer.framePrefix);
				Frame currFrame = (Frame) context.getPointers().getVariable(
						tokens[2]).getValue();
				// Create the ip variable
				for (Item i : currFrame.getItems()) {
					context.getPointers().setObject(tokens[1], i);
					Status status = RunFrameAndReportError(code, context);
					// check if we need to exit this loop because of
					// statements in
					// the code that was run
					if (status == Status.Exit || status == Status.Return)
						return status;
					else if (status == Status.Break)
						return Status.OK;
				}
				return Status.OK;
			}

			if (Primitives.isPrimitive(tokens[1])
					&& Primitives.isPrimitive(tokens[2])) {
				if (tokens[0].equals("movecursorto")) {
					assertExactParametreCount(tokens, 2);
					int x = (int) context.getPrimitives().getIntegerValue(
							tokens[1]);
					int y = (int) context.getPrimitives().getIntegerValue(
							tokens[2]);
					DisplayIO.setCursorPosition(x, y);
					return Status.OK;
				} else if (tokens[0].equals("not")) {
					context.getPrimitives().not(tokens[1], tokens[2]);
				} else if (tokens[0].equals("exp")) {
					context.getPrimitives().exp(tokens[1], tokens[2]);
				} else if (tokens[0].equals("log")) {
					context.getPrimitives().log(tokens[1], tokens[2]);
				} else if (tokens[0].equals("log10")) {
					context.getPrimitives().log10(tokens[1], tokens[2]);
				} else if (tokens[0].equals("sqrt")) {
					context.getPrimitives().sqrt(tokens[1], tokens[2]);
				} else if (tokens[0].equals("add")) {
					context.getPrimitives().add(tokens[1], tokens[2]);
				} else if (tokens[0].equals("subtract")) {
					context.getPrimitives().subtract(tokens[1], tokens[2]);
				} else if (tokens[0].equals("multiply")) {
					context.getPrimitives().multiply(tokens[1], tokens[2]);
				} else if (tokens[0].equals("divide")) {
					context.getPrimitives().divide(tokens[1], tokens[2]);
				} else
					throw new RuntimeException("Invalid statement:");
				return Status.OK;
			}
		} else if (tokens.length == 4) {
			if (Primitives.isPrimitive(tokens[1])
					&& Primitives.isPrimitive(tokens[2])
					&& Primitives.isPrimitive(tokens[3])) {
				if (tokens[0].equals("add")) {
					context.getPrimitives()
							.add(tokens[1], tokens[2], tokens[3]);
				} else if (tokens[0].equals("subtract")) {
					context.getPrimitives().subtract(tokens[1], tokens[2],
							tokens[3]);
				} else if (tokens[0].equals("divide")) {
					context.getPrimitives().divide(tokens[1], tokens[2],
							tokens[3]);
				} else if (tokens[0].equals("multiply")) {
					context.getPrimitives().multiply(tokens[1], tokens[2],
							tokens[3]);
				} else if (tokens[0].equals("modulo")) {
					context.getPrimitives().modulo(tokens[1], tokens[2],
							tokens[3]);
				} else if (tokens[0].equals("power")) {
					context.getPrimitives().power(tokens[1], tokens[2],
							tokens[3]);
				} else
					throw new Exception("Statement not found: " + tokens[0]
							+ " in " + code.toString());
				return Status.OK;
			}
		}
		throw new RuntimeException("Unknown statement: ");
	}

	public static int countCharsInString(String s, String pattern) {
		String newString = s;
		int count = -1;
		do {
			count++;
			s = newString;
			newString = s.replaceFirst(pattern, "");
		} while (s.length() != newString.length());

		return count;
	}

	public static void assertVariableType(String varName, int no, String type)
			throws Exception {
		if (!varName.startsWith(type))
			throw new IncorrectTypeException(type, no);
	}

	/*
	 * public static void assertPrimitiveType(String varName, int no) throws
	 * Exception { if (!Primitives.isPrimitive(varName)) throw new
	 * IncorrectTypeException("primitive", no); }
	 */

	public static void assertMinParametreCount(String[] tokens,
			int minParametres) throws Exception {
		if (tokens.length - 1 < minParametres)
			throw new BelowMinParametreCountException(minParametres);
	}

	public static void assertExactParametreCount(String[] tokens,
			int parametreCount) throws Exception {
		if (tokens.length - 1 != parametreCount)
			throw new IncorrectParametreCountException(parametreCount);
	}
}