package org.expeditee.gui;

import java.awt.Color;
import java.awt.Polygon;
import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import org.expeditee.io.Logger;
import org.expeditee.items.Dot;
import org.expeditee.items.Item;
import org.expeditee.items.ItemPermission;
import org.expeditee.items.ItemUtils;
import org.expeditee.items.Line;
import org.expeditee.items.Picture;
import org.expeditee.items.Text;

public class FrameUtils {

	private static final int COLUMN_WIDTH = 50;

	/**
	 * Provides a way to monitor the time elapsed between button-down and the
	 * finished painting.
	 */
	public static TimeKeeper ResponseTimer = new TimeKeeper();

	private static float _ResponseTimeSum = 0;

	private static float _LastResponse = 0;

	public static Item LastEdited = null;

	public static int MINIMUM_INTERITEM_SPACING = -6;

	public static float getResponseTimeTotal() {
		return _ResponseTimeSum;
	}

	public static float getLastResponseTime() {
		return _LastResponse;
	}

	/**
	 * Checks if the given top Item is above the given bottom Item, allowing for
	 * the X coordinates to be off by a certain width...
	 * 
	 * @param item1
	 *            The Item to check is above the other Item
	 * @param item2
	 *            The Item to check is below the top Item
	 * @return True if top is above bottom, False otherwise.
	 */
	public static boolean inSameColumn(Item item1, Item item2) {
		if (!(item1 instanceof Text) || !(item2 instanceof Text))
			return false;

		if (item1.getID() < 0 || item2.getID() < 0)
			return false;

		int minX = item2.getX();
		int maxX = item2.getX() + item2.getBoundsWidth();

		int startX = item1.getX();
		int endX = item1.getX() + item1.getBoundsWidth();

		// Check that the two items left values are close
		if (Math.abs(item1.getX() - item2.getX()) > COLUMN_WIDTH)
			return false;

		// Ensure the two items
		if ((minX >= startX && minX <= endX)
				|| (maxX >= startX && maxX <= endX)
				|| (startX >= minX && startX <= maxX)
				|| (endX >= minX && endX <= maxX))
			return true;

		return false;
	}

	public static boolean sameBulletType(String bullet1, String bullet2) {
		if (bullet1 == null || bullet2 == null)
			return false;

		if (bullet1 == "" || bullet2 == "")
			return false;

		if (Character.isLetter(bullet1.charAt(0))
				&& Character.isLetter(bullet2.charAt(0)))
			return true;

		if (Character.isDigit(bullet1.charAt(0))
				&& Character.isDigit(bullet2.charAt(0)))
			return true;

		// TODO make this more sofisticated

		return false;
	}

	private static boolean needsRenumbering(String s) {
		if (s == null || s == "")
			return false;
		if (!Character.isLetterOrDigit(s.charAt(0)))
			return false;

		s = s.trim();
		// if its all letters then we dont want to auto adjust
		if (s.length() > 2) {
			for (int i = 0; i < s.length() - 1; i++) {
				if (!Character.isLetter(s.charAt(i)))
					return true;
			}
		} else
			return true;

		return false;
	}

	/**
	 * 
	 * @param toAlign
	 * @param moveAll
	 * @param adjust
	 * @return
	 */
	public static int Align(List<Item> toAlign, boolean moveAll, int adjust) {
		Collections.sort(toAlign);

		// Single items dont need alignment
		if (toAlign.size() < 2)
			return 0;

		// get the first item
		Item from = toAlign.get(0);
		if (from.getParent() == null)
			from = toAlign.get(1);
		int x = from.getX();

		Frame curr = from.getParent();
		Item above = curr.getItemAbove(from);

		String lastBullet = "";

		if (above != null && curr.isNormalTextItem(above))
			lastBullet = FrameKeyboardActions.getAutoBullet(((Text) above)
					.getTextNoList());
		else {
			lastBullet = FrameKeyboardActions.getBullet(((Text) toAlign.get(0))
					.getTextNoList());
		}
		if (needsRenumbering(lastBullet)) {
			// renumber...
			for (int i = 0; i < toAlign.size(); i++) {
				assert toAlign.get(i) instanceof Text;

				Text currentText = ((Text) toAlign.get(i));
				String currentBullet = FrameKeyboardActions
						.getAutoBullet(currentText.getTextNoList());

				if (sameBulletType(lastBullet, currentBullet)) {
					currentText.stripFirstWord();

					currentText.setText(lastBullet
							+ currentText.getTextNoList());
					lastBullet = FrameKeyboardActions.getAutoBullet(currentText
							.getTextNoList());
				}
			}
		}

		// work out the spacing between the first item and the one above it

		int space = 10 + adjust;

		// if we are dropping from the title make the space a little bigger
		// than normal
		if (above != curr.getTitle() && above != null) {
			// Make the gap between all items the same as the gap between
			// the first two
			space = (int) (from.getPolygon().getBounds().getMinY() - above
					.getPolygon().getBounds().getMaxY());

			if (space < MINIMUM_INTERITEM_SPACING)
				space = MINIMUM_INTERITEM_SPACING;

			if (above != curr.getName() && above != curr.getTitle())
				x = above.getX();

			space += adjust;

			from.setY((int) above.getPolygon().getBounds().getMaxY()
					+ space
					+ ((int) (from.getY() - from.getPolygon().getBounds()
							.getMinY())));

			if (moveAll)
				from.setX(x);
		}

		for (int i = 1; i < toAlign.size(); i++) {
			Item current = toAlign.get(i);
			Item top = toAlign.get(i - 1);

			// The bottom of the previous item
			int bottom = (int) top.getPolygon().getBounds().getMaxY();

			// the difference between the current item's Y coordinate and
			// the top of the highlight box
			int diff = (int) (current.getY() - current.getPolygon().getBounds()
					.getMinY());

			int newPos = bottom + space + diff;

			if (moveAll) {
				current.setPosition(x, newPos);
			} else if (newPos > current.getY()) {
				current.setY(newPos);
			}

		}

		// if (insert != null)
		// return insert.getY();

		// Michael thinks we return the y value for the next new item??
		int y = from.getY() + from.getBoundsHeight() + space;
		return y;
	}

	public static boolean LeavingFrame(Frame current) {
		// active overlay frames may also require saving if they have been
		// changed
		for (Overlay o : current.getOverlays())
			if (!SaveCheck(o.Frame))
				return false;

		// if the check fails there is no point continuing
		if (!SaveCheck(current))
			return false;

		for (Item i : current.getItems())
			i.showHighlight(false);
		return true;
	}

	private static boolean SaveCheck(Frame toSave) {
		// don't bother saving frames that haven't changed
		if (!toSave.hasChanged())
			return true;

		// if the frame has been changed, then save it
		if (DisplayIO.isTwinFramesOn()) {
			Frame opposite = DisplayIO.getOppositeFrame();

			String side = "left";
			if (DisplayIO.getCurrentSide() == 0)
				side = "right";

			// if the two frames both have changes, prompt the user for the
			// next
			// move
			if (opposite.hasChanged() && opposite.equals(toSave)) {
				if (DisplayIO.DisplayConfirmDialog(
						"Leaving this frame will discard changes made in the "
								+ side + " Frame. Continue?", "Changes",
						DisplayIO.TYPE_WARNING, DisplayIO.OPTIONS_OK_CANCEL,
						DisplayIO.RESULT_OK)) {
					FrameIO.SaveFrame(toSave);
					DisplayIO.Reload(DisplayIO.FrameOnSide(opposite));
					return true;
				} else
					return false;
			} else if (opposite.getOverlays().contains(new Overlay(toSave, 1))) {
				if (opposite.getOverlays().get(
						opposite.getOverlays().indexOf(new Overlay(toSave, 1))).Frame
						.hasChanged())
					if (DisplayIO.DisplayConfirmDialog(
							"Leaving this frame will discard changes made in the "
									+ side + " Frame. Continue?", "Changes",
							DisplayIO.TYPE_WARNING,
							DisplayIO.OPTIONS_OK_CANCEL, DisplayIO.RESULT_OK)) {
						FrameIO.SaveFrame(toSave);
						DisplayIO.Reload(DisplayIO.FrameOnSide(opposite));
						return true;
					} else
						return false;
			}

			// save the current frame and restore the other side
			FrameIO.SaveFrame(toSave);
			return true;
		}

		// single-frame mode can just save and return
		FrameIO.SaveFrame(toSave);
		return true;
	}

	/**
	 * Displays the given Frame on the display. If the current frame has changed
	 * since the last save then it will be saved before the switch is made. The
	 * caller can also dictate whether the current frame is added to the
	 * back-stack or not.
	 * 
	 * @param toDisplay
	 *            The Frame to display on the screen
	 * @param addToBack
	 *            True if the current Frame should be added to the back-stack,
	 *            False otherwise
	 */
	public static void DisplayFrame(Frame toDisplay, boolean addToBack) {
		if (toDisplay == null)
			return;

		Frame current = DisplayIO.getCurrentFrame();

		// move any anchored connected items
		if (Frame.itemAttachedToCursor()) {
			ArrayList<Item> toAdd = new ArrayList<Item>();

			for (Item i : Frame.FreeItems) {
				for (Item x : i.getConnected())
					if (!Frame.FreeItems.contains(x) && !toAdd.contains(x))
						toAdd.add(x);
			}

			current.removeAllItems(toAdd);

			boolean oldChange = toDisplay.hasChanged();
			toDisplay.updateIDs(toAdd);
			toDisplay.addAllItems(toAdd);
			toDisplay.setChanged(oldChange);
		}

		// if the saving happened properly, we can continue
		if (!LeavingFrame(current)) {
			FrameGraphics.DisplayMessage("Navigation cancelled");
			return;
		}

		if (addToBack && current != toDisplay)
			DisplayIO.addToBack(current);

		Parse(toDisplay);
		DisplayIO.setCurrentFrame(toDisplay);

		// update response timer
		_LastResponse = ResponseTimer.getElapsedSeconds();
		_ResponseTimeSum += _LastResponse;
	}

	/**
	 * Loads and displays the Frame with the given framename, and adds the
	 * current frame to the back-stack if required.
	 * 
	 * @param framename
	 *            The name of the Frame to load and display
	 * @param addToBack
	 *            True if the current Frame should be added to the back-stack,
	 *            false otherwise
	 */
	public static void DisplayFrame(String frameName, boolean addToBack) {
		Frame newFrame = getFrame(frameName);

		if (newFrame != null)
			// display the frame
			DisplayFrame(newFrame, addToBack);
	}

	/**
	 * Loads and displays the Frame with the given framename and adds the
	 * current frame to the back-stack. This is the same as calling
	 * DisplayFrame(framename, true)
	 * 
	 * @param framename
	 *            The name of the Frame to load and display
	 */
	public static void DisplayFrame(String framename) {
		DisplayFrame(framename, true);
	}

	public static void DisplayHomeFrame() {
		DisplayFrame(UserSettings.FirstFrame);
	}

	public static void DisplayProfileFrame() {
		DisplayFrame(UserSettings.Username
				+ FrameIO.getFramesetExtension(UserSettings.Username));
	}

	public static Frame getFrame(String frameName) {
		// if the new frame does not exist then tell the user
		Frame f = FrameIO.LoadFrame(frameName);

		if (f == null) {
			FrameGraphics.ErrorMessage("Frame '" + frameName
					+ "' could not be found.");
		}

		return f;
	}

	/**
	 * Creates a new Picture Item from the given Text source Item and adds it to
	 * the given Frame.
	 * 
	 * @return True if the image was created successfully, false otherwise
	 */
	private static boolean createPicture(Frame frame, Text txt) {
		// attempt to create the picture
		Picture pic = ItemUtils.CreatePicture(txt, frame);

		// if the picture could not be created successfully
		if (pic == null) {
			assert (txt.getFirstLine() != null);
			assert (txt.getFirstLine().split(" ").length > 1);

			FrameGraphics
					.ErrorMessage("Image " + txt.getFirstLine().split(" ")[1]
							+ " could not be loaded");
			return false;
		}

		// replace the Text source with the new Picture
		frame.removeItem(txt);
		frame.addItem(pic);

		return true;
	}

	private static boolean createFramePicture(Frame frame, Text txt) {
		if (txt.getLink() == null)
			return false;

		// attempt to create the picture
		Picture pic = ItemUtils.CreateFramePicture(txt, frame);

		// if the picture could not be created successfully
		if (pic == null) {
			FrameGraphics.ErrorMessage("Error loading frame image");
			return false;
		}

		// replace the Text source with the new Picture
		frame.removeItem(txt);
		frame.addItem(pic);

		return true;
	}

	public static void ParseProfile(Frame profile) {
		if (profile == null)
			return;

		UserSettings.TitleTemplate = profile.getTitle();

		List<Item> items = profile.getItems();

		// check for all tags setting user values
		for (Item item : items) {
			if (ItemUtils.isTag(item, "@HomeFrame:")) {
				String first = getLink(item, UserSettings.FirstFrame);
				// do not use non-existant frames as the first frame
				if (FrameIO.isValidFrameName(first)) {
					UserSettings.FirstFrame = first;
				}
				// warn the user
				else {
					FrameGraphics.WarningMessage("Home frame: " + first
							+ " could not be found.");
					UserSettings.FirstFrame = profile.getFrameName();
				}
			} else if (ItemUtils.isTag(item, "@MenuFrame:"))
				UserSettings.MenuFrame = getLink(item, UserSettings.MenuFrame);
			else if (ItemUtils.isTag(item, "@DefaultFrame:"))
				UserSettings.DefaultFrame = getLink(item,
						UserSettings.DefaultFrame);
			else if (ItemUtils.isTag(item, "@AntiAlias:"))
				UserSettings.AntiAlias = getBoolean(item,
						UserSettings.AntiAlias);
			else if (ItemUtils.isTag(item, "@Gravity:"))
				UserSettings.Gravity = getInt(item, UserSettings.Gravity);
			else if (ItemUtils.isTag(item, "@ShowLineHighlight:"))
				UserSettings.LineHighlight = getBoolean(item,
						UserSettings.LineHighlight);
			else if (ItemUtils.isTag(item, "@LineStraightenThreshold:"))
				UserSettings.LineStraightenThreshold = getInt(item,
						UserSettings.LineStraightenThreshold);
			else if (ItemUtils.isTag(item, "@NoOpThreshold:"))
				UserSettings.NoOpThreshold = getInt(item,
						UserSettings.NoOpThreshold);
			else if (ItemUtils.isTag(item, "@InitialWidth:")){
				UserSettings.InitialWidth= getInt(item,
						UserSettings.InitialWidth);}
			else if (ItemUtils.isTag(item, "@InitialHeight:")){
				UserSettings.InitialHeight= getInt(item,
						UserSettings.InitialHeight);}
			else if (ItemUtils.isTag(item, "@Logging:")) {
				UserSettings.Logging = getBoolean(item, UserSettings.Logging);
			} else if (ItemUtils.isTag(item, ItemUtils.TAG_ITEM_TEMPLATE)) {
				UserSettings.ItemTemplate = ((Text) item).getTemplateForm();
			} else if (ItemUtils.isTag(item, ItemUtils.TAG_ANNOTATION_TEMPLATE)) {
				UserSettings.AnnotationTemplate = ((Text) item)
						.getTemplateForm();
			} else if (ItemUtils.isTag(item,
					ItemUtils.TAG_CODE_COMMENT_TEMPLATE)) {
				UserSettings.CodeCommentTemplate = ((Text) item)
						.getTemplateForm();
			} else if (ItemUtils.isTag(item, "@FrameDir:")) {
				String dir = getDir(item, null);
				if (dir != null)
					UserSettings.FrameDirs.add(0, dir);
			} else if (ItemUtils.isTag(item, "@LogDir:")) {
				org.expeditee.gui.FrameIO.LOGS_DIR = getDir(item,
						org.expeditee.gui.FrameIO.LOGS_DIR);
			} else if (ItemUtils.isTag(item, "@ImageDir:")) {
				String dir = getDir(item, null);
				if (dir != null)
					UserSettings.ImageDirs.add(0, dir);
			} else if (ItemUtils.isTag(item, "@Threading:")) {
				UserSettings.Threading = getBoolean(item,
						UserSettings.Threading);
			}
		}

		if (UserSettings.FirstFrame == null)
			UserSettings.FirstFrame = profile.getFrameName();

	}

	private static String getLink(Item item, String alt) {
		if (item == null || !(item instanceof Text))
			return alt;

		String value = AttributeUtils.stripValue(((Text) item).getFirstLine())
				.trim().toLowerCase();
		if (value.length() > 0) {
			item.setLink(value);
			return value;
		} else if (item.getLink() != null)
			return item.getLink();

		return alt;
	}

	private static boolean getBoolean(Item item, boolean alt) {
		if (item == null || !(item instanceof Text))
			return alt;

		String value = AttributeUtils.stripValue(((Text) item).getFirstLine())
				.trim().toLowerCase();
		if (value != null && value.length() > 0) {
			if (value.equals("t") || value.equals("true")
					|| value.equals("yes") || value.equals("y"))
				return true;

			if (value.equals("f") || value.equals("false")
					|| value.equals("no") || value.equals("n")) {
				return false;
			}
		}

		return alt;
	}

	private static int getInt(Item item, int alt) {
		if (item == null || !(item instanceof Text))
			return alt;

		String value = AttributeUtils.stripValue(((Text) item).getFirstLine())
				.trim().toLowerCase();
		try {
			return Integer.parseInt(value);
		} catch (Exception e) {
		}

		return alt;
	}

	private static String getDir(Item item, String alt) {
		String name = getString(item, null);
		if (name != null) {
			File tester = new File(name);
			if (tester.exists() && tester.isDirectory()) {
				if (name.endsWith(File.separator))
					return name;
				else
					return name + File.separator;
			}
		}

		return alt;
	}

	/*
	 * private static String FindFile(List<Item> items, String tag, String
	 * alt){ String name = FindString(items, tag, null); if(name != null){ File
	 * tester = new File(name); if(tester.exists()) return name; }
	 * 
	 * return alt; }
	 */
	private static String getString(Item item, String alt) {
		if (item == null || !(item instanceof Text))
			return alt;

		String value = AttributeUtils.stripValue(((Text) item).getFirstLine())
				.trim().toLowerCase();
		if (value != null && value.length() > 0)
			return value;

		return alt;
	}

	public static void Parse(Frame toParse) {
		Parse(toParse, false);
	}

	/**
	 * Checks for any special Annotation items and updates the display as
	 * necessary. Special Items: Images, overlays, sort.
	 * 
	 */
	public static void Parse(Frame toParse, boolean firstParse) {
		// System.out.println(firstParse);
		if (firstParse)
			ItemUtils.EnclosedCheck(toParse.getItems());

		ArrayList<Item> items = new ArrayList<Item>();
		items.addAll(toParse.getItems());

		// if XRayMode is on, replace pictures with their underlying text
		if (FrameGraphics.isXRayMode()) {
			for (Item i : items) {
				if (i instanceof Picture) {
					toParse.removeItem(i);
					toParse.addItem(((Picture) i).getText());
				}
			}

			// update the list of items
			items.clear();
			items.addAll(toParse.getItems());
		}

		// Text title = null;
		// Text template = UserSettings.ItemTemplate.copy();

		ArrayList<Overlay> overlays = new ArrayList<Overlay>();

		// disable reading of cached overlays if in twinframes mode
		if (DisplayIO.isTwinFramesOn())
			FrameIO.SuspendCache();

		int pointtype = Item.POINTTYPE_SQUARE;
		boolean filledPoints = true;

		// check for any new overlay items
		for (Item i : items) {
			try {
				/*
				 * if (title == null && i.isTitle()) { title = (Text) i;
				 * 
				 * if (title.getX() <= 0 && title.getY() <= 0)
				 * title.resetTitlePosition(); }
				 */

				// check for pointtype settings
				if (ItemUtils.isTag(i, ItemUtils.TAG_POINTTYPE)) {
					Text txt = (Text) i;
					String line = txt.getFirstLine();
					line = ItemUtils.StripTag(line, ItemUtils
							.GetTag(ItemUtils.TAG_POINTTYPE));

					if (line != null) {
						line = line.toLowerCase();
						if (line.indexOf(" ") > 0) {
							String fill = line.substring(line.indexOf(" ") + 1);
							if (fill.startsWith("nofill"))
								filledPoints = false;
							else
								filledPoints = true;
						}

						if (line.startsWith("circle"))
							pointtype = Item.POINTTYPE_CIRCLE;
						else
							pointtype = Item.POINTTYPE_SQUARE;
					}
				}

				// check for Images
				if (!FrameGraphics.isXRayMode()) {
					if (ItemUtils.isTag(i, ItemUtils.TAG_IMAGE))
						createPicture(toParse, (Text) i);
					// check for frame images
					else if (i.getLink() != null
							&& ItemUtils.isTag(i, ItemUtils.TAG_FRAME_IMAGE))
						createFramePicture(toParse, (Text) i);
				}

				// check for new overlay items
				if (ItemUtils.isTag(i, ItemUtils.TAG_OVERLAY)
						&& i.getLink() != null) {
					Frame overlay = FrameIO.LoadFrame(i.getLink());
					// Parse(overlay);
					if (overlay != null && !overlays.contains(overlay))
						overlays.add(new Overlay(overlay,
								ItemPermission.PERMISSION_NONE));
				}

				// check for new overlay items
				if (ItemUtils.isTag(i, ItemUtils.TAG_ACTIVE_OVERLAY)
						&& i.getLink() != null) {
					String link = i.getLink();
					if (FrameIO.isPositiveInteger(link))
						link = toParse.getFramesetNameAdjusted()
								+ link;
					Frame overlay = FrameIO.LoadFrame(link);
					// Parse(overlay);
					if (overlay != null && !overlays.contains(overlay)) {
						// get level if specified
						String level = ItemUtils.StripTag(((Text) i)
								.getFirstLine());
						// default permission (if none is specified)
						int perm = ItemPermission.PERMISSION_FOLLOW_LINKS;

						try {
							perm = Integer.parseInt(level);
						} catch (Exception e) {
							// if this is a text permission level, parse it
							if (level.length() > 0) {
								level = level.toLowerCase().trim();
								if (level.equals("none"))
									perm = ItemPermission.PERMISSION_NONE;
								else if (level.equals("links")
										|| level.equals("link"))
									perm = ItemPermission.PERMISSION_FOLLOW_LINKS;
								else if (level.equals("copy"))
									perm = ItemPermission.PERMISSION_COPY;
								else if (level.equals("tdfc"))
									perm = ItemPermission.PERMISSION_TDFC;
								else if (level.equals("full")
										|| level.equals("all"))
									perm = ItemPermission.PERMISSION_FULL;
								else
									FrameGraphics
											.DisplayMessage("Permission level '"
													+ level
													+ "' not recognized, using default level (Links)");
							}
						}

						overlays.add(new Overlay(overlay, perm));
					}
				}

				/*
				 * check for Item templates if (ItemUtils.isTag(i,
				 * ItemUtils.TAG_ITEM_TEMPLATE)) { template = (Text) i; }
				 */

				// only remove items on first parse
				if (firstParse) {
					// remove dots of size 1
					if (i instanceof Dot) {
						Dot dot = (Dot) i;
						if (dot.getSize() <= 1
								&& (dot.getLines() == null || dot.getLines()
										.size() == 0))
							if (dot.getConstraints() == null
									|| dot.getConstraints().size() == 0) {
								toParse.removeItem(dot);
							}
					}

					// remove Text that is all whitespace
					if (i instanceof Text) {
						List<String> text = ((Text) i).getText();
						// remove empty text items
						if (text == null || text.size() == 0)
							toParse.removeItem(i);
						else {
							for (String s : text)
								if (s.trim().length() > 0)
									break;
								else
									toParse.removeItem(i);
						}
					}

				}

				if (toParse.getFrameNumber() > 0
						&& ItemUtils.isTag(i, ItemUtils.TAG_PARENT)
						&& i.getLink() == null)
					i.setLink(DisplayIO.getCurrentFrame().getFrameName());

			} catch (Exception e) {
				Logger.Log(e);
				e.printStackTrace();
				FrameGraphics.WarningMessage("Exception occured when loading "
						+ i.getClass().getSimpleName() + "(ID: " + i.getID()
						+ ") " + e.getMessage());
			}
		}

		for (Item i : items) {
			if (i instanceof Dot) {
				((Dot) i).setPointType(pointtype);
				((Dot) i).useFilledPoints(filledPoints);
			}
		}

		FrameIO.ResumeCache();

		synchronized (toParse) {
			toParse.clearOverlays();
			toParse.addAllOverlays(overlays);
		}
	}

	/**
	 * Searches through the list of items on this frame to find one at the given
	 * x,y coordinates.
	 * 
	 * @param x
	 *            The x coordinate
	 * @param y
	 *            The y coordinate
	 * @return The Item at the given coordinates, or NULL if none is found.
	 */
	private static ItemPermission onItem(Frame toCheck, int x, int y) {
		if (toCheck == null)
			return null;

		List<ItemPermission> possibles = new ArrayList<ItemPermission>(0);

		// if the mouse is in the message area
		if (y > FrameGraphics.getMaxFrameSize().getHeight()) {
			// check the individual message items
			for (Text message : FrameGraphics.Messages) {
				if (message != null && message.contains(x, y))
					possibles.add(new ItemPermission(message,
							ItemPermission.PERMISSION_COPY));
			}

			// check the link to the message frame
			if (FrameGraphics.MessageLink != null) {
				if (FrameGraphics.MessageLink.contains(x, y))
					possibles.add(new ItemPermission(FrameGraphics.MessageLink,
							ItemPermission.PERMISSION_COPY));
			}
			// otherwise, the mouse is on the frame
		} else {

			if (LastEdited != null && LastEdited.contains(x, y)
					&& !Frame.FreeItems.contains(LastEdited)
					&& LastEdited.getParent() == DisplayIO.getCurrentFrame()
					&& LastEdited.getParent().getItems().contains(LastEdited))
				return new ItemPermission(LastEdited,
						ItemPermission.PERMISSION_FULL);
			else
				LastEdited = null;

			ArrayList<Item> checkList = new ArrayList<Item>();
			checkList.addAll(toCheck.getItems());
			for (Item i : checkList) {
				// do not check annotation items in audience mode
				if (!(FrameGraphics.isAudienceMode() && i.isAnnotation())) {
					if (i.contains(x, y) && !Frame.FreeItems.contains(i)) {
						// names have copy permissions only
						if (i == toCheck.getName())
							possibles.add(new ItemPermission(i,
									ItemPermission.PERMISSION_TDFC));
						else
							possibles.add(new ItemPermission(i,
									ItemPermission.PERMISSION_FULL));
					}
				}
			}

			for (Overlay o : toCheck.getOverlays()) {
				if (o.Level > ItemPermission.PERMISSION_NONE) {
					ItemPermission i = onItem(o.Frame, x, y);

					// the message is included above with full permissions,
					// it
					// should be ignored here
					if (i != null
							&& /* i.Item != FrameGraphics.TextMessage && */i.Item
									.getID() > 0) {
						if (o.Level < i.Permission)
							i.Permission = o.Level;
						possibles.add(i);
					}
				}
			}
		}

		// if there is only one possibility, return it
		if (possibles.size() == 1)
			return possibles.get(0);

		// if there are no possible items, return null
		if (possibles.size() == 0)
			return null;

		// return closest x,y pair to mouse
		ItemPermission closest = possibles.get(0);
		int distance = (int) Math.sqrt(Math.pow(Math.abs(closest.Item.getX()
				- x), 2)
				+ Math.pow(Math.abs(closest.Item.getY() - y), 2));

		for (ItemPermission ip : possibles) {
			Item i = ip.Item;

			int d = (int) Math.sqrt(Math.pow(Math.abs(i.getX() - x), 2)
					+ Math.pow(Math.abs(i.getY() - y), 2));

			if (d < distance) {
				distance = d;

				// dots take precedence over lines
				/**
				 * TODO: Remove line\dot specification
				 */
				if ((!(closest.Item instanceof Dot && i instanceof Line))
						&& (!(closest.Item instanceof Text && i instanceof Line)))
					closest = ip;

			}

		}

		return closest;
	}

	public static ItemPermission getCurrentItem() {
		return onItem(DisplayIO.getCurrentFrame(), DisplayIO.getMouseX(),
				DisplayIO.getMouseY());
	}

	public static List<ItemPermission> getCurrentItems() {
		/*
		 * // update enclosed shapes Frame current =
		 * DisplayIO.getCurrentFrame(); List<Item> items = current.getItems();
		 * ArrayList<Dot> dots = new ArrayList<Dot>(0);
		 * 
		 * for (Item i : items) if (i instanceof Dot) dots.add((Dot) i);
		 * 
		 * ArrayList<Polygon> possibles = new ArrayList<Polygon>(0); ArrayList<Dot>
		 * used = new ArrayList<Dot>(0);
		 * 
		 * for (Dot d : dots) if (d.isEnclosed()) { boolean skip = false; //
		 * check for duplicate enclosures for (Dot seen : used) { if
		 * (seen.getEnclosingDots().contains(d)) { skip = true; break; } }
		 * 
		 * if (!skip) { Polygon p = d.getEnclosedShape(); if (p
		 * .contains(DisplayIO.getMouseX(), DisplayIO .getMouseY())) {
		 * possibles.add(p); used.add(d); } } }
		 * 
		 * if (possibles.size() == 0) return null; // if there is only one
		 * possibility, return it if (possibles.size() == 1) { return
		 * getItemsEnclosedBy(current, possibles.get(0)); //otherwise, determine
		 * which polygon is closest to the cursor }else{
		 * Collections.sort(possibles, new Comparator<Polygon>(){ public int
		 * compare(Polygon p1, Polygon p2) { int closest = Integer.MAX_VALUE;
		 * int close2 = Integer.MAX_VALUE;
		 * 
		 * int mouseX = DisplayIO.getMouseX(); int mouseY =
		 * DisplayIO.getMouseY();
		 * 
		 * for(int i = 0; i < p1.npoints; i++){ int diff =
		 * Math.abs(p1.xpoints[i] - mouseX) + Math.abs(p1.ypoints[i] - mouseY);
		 * int diff2 = Integer.MAX_VALUE;
		 * 
		 * if(i < p2.npoints) diff2 = Math.abs(p2.xpoints[i] - mouseX) +
		 * Math.abs(p2.ypoints[i] - mouseY);
		 * 
		 * if(diff < Math.abs(closest)){ close2 = closest; closest = diff; }else
		 * if(diff < Math.abs(close2)) close2 = diff;
		 * 
		 * if(diff2 < Math.abs(closest)){ close2 = closest; closest = -diff2;
		 * }else if(diff2 < Math.abs(close2)) close2 = diff2; }
		 * 
		 * if(closest > 0 && close2 > 0) return -10;
		 * 
		 * if(closest < 0 && close2 < 0) return 10;
		 * 
		 * 
		 * if(closest > 0) return -10;
		 * 
		 * return 10; }
		 * 
		 * }); return getItemsEnclosedBy(current, possibles.get(0)); }
		 */

		List<Dot> enclosure = getEnclosingDots();
		if (enclosure == null || enclosure.size() == 0)
			return null;

		return getItemsEnclosedBy(DisplayIO.getCurrentFrame(), enclosure.get(0)
				.getEnclosedShape());
	}

	public static List<Dot> getEnclosingDots() {
		// update enclosed shapes
		Frame current = DisplayIO.getCurrentFrame();
		List<Item> items = current.getItems();

		ArrayList<Dot> used = new ArrayList<Dot>(0);
		ArrayList<Dot> seen = new ArrayList<Dot>(0);

		for (Item i : items)
			if (i instanceof Dot) {
				Dot d = (Dot) i;
				if (d.isEnclosed() && !seen.contains(d)) {
					Polygon p = d.getEnclosedShape();
					if (p
							.contains(DisplayIO.getMouseX(), DisplayIO
									.getMouseY())) {

						used.add(d);
						seen.addAll(d.getEnclosingDots());
					}
				}
			}

		if (used.size() == 0)
			return null;

		// if there is only one possibility, return it
		if (used.size() == 1) {
			return used.get(0).getEnclosingDots();
			// otherwise, determine which polygon is closest to the cursor
		} else {
			Collections.sort(used, new Comparator<Dot>() {
				public int compare(Dot d1, Dot d2) {
					Polygon p1 = d1.getEnclosedShape();
					Polygon p2 = d2.getEnclosedShape();

					int closest = Integer.MAX_VALUE;
					int close2 = Integer.MAX_VALUE;

					int mouseX = DisplayIO.getMouseX();
					int mouseY = DisplayIO.getMouseY();

					for (int i = 0; i < p1.npoints; i++) {
						int diff = Math.abs(p1.xpoints[i] - mouseX)
								+ Math.abs(p1.ypoints[i] - mouseY);
						int diff2 = Integer.MAX_VALUE;

						if (i < p2.npoints)
							diff2 = Math.abs(p2.xpoints[i] - mouseX)
									+ Math.abs(p2.ypoints[i] - mouseY);

						if (diff < Math.abs(closest)) {
							close2 = closest;
							closest = diff;
						} else if (diff < Math.abs(close2))
							close2 = diff;

						if (diff2 < Math.abs(closest)) {
							close2 = closest;
							closest = -diff2;
						} else if (diff2 < Math.abs(close2))
							close2 = diff2;
					}

					if (closest > 0 && close2 > 0)
						return -10;

					if (closest < 0 && close2 < 0)
						return 10;

					if (closest > 0)
						return -10;

					return 10;
				}

			});

			return used.get(0).getEnclosingDots();
		}
	}

	public static List<ItemPermission> getItemsEnclosedBy(Frame frame,
			Polygon poly) {
		List<Item> contained = frame.getItemsWithin(poly);

		ArrayList<ItemPermission> results = new ArrayList<ItemPermission>(
				contained.size());

		// check for correct permissions
		for (Item item : contained) {
			// if the item is on the frame
			if (item.getParent() == frame || item.getParent() == null)
				// TODO this is a bit of a hack...
				// the
				// item
				// parent
				// should
				// really
				// be
				// set
				// when
				// the
				// polygon
				// is
				// created...
				results.add(new ItemPermission(item,
						ItemPermission.PERMISSION_FULL));
			// otherwise, it must be on an overlay frame
			else {
				for (Overlay overlay : frame.getOverlays())
					if (overlay.Frame == item.getParent()) {
						results.add(new ItemPermission(item, overlay.Level));
						break;
					}
			}

		}

		return results;
	}

	/**
	 * Fills the given Frame with default profile tags
	 */
	public static void CreateDefaultProfile(Frame profile) {
		Text title = profile.getTitle();
		title.setText("Profile Frame");
		title.setSize(50);
		title.setFontStyle("Bold");
		title.setFamily("SansSerif");
		title.setColor(Color.BLUE);
		title.setPosition(25, 50);

		int spacing = 50;
		final int intialYPos = 75;
		int xPos = 75;
		int yPos = intialYPos;

		yPos += spacing;
		profile.addText(xPos, yPos, "@HomeFrame: " + profile.getFrameName(),
				null, profile.getFrameName());
		yPos += spacing;
		String defaultFrameName = profile.getFramesetNameAdjusted() + "0";
		profile.addText(xPos, yPos, "@DefaultFrame: " + defaultFrameName, null,
				defaultFrameName);
		yPos += spacing;

		profile.addText(xPos, yPos, "@InitialWidth: " + UserSettings.InitialWidth, null);
		yPos += spacing;

		profile.addText(xPos, yPos, "@InitialHeight: " + UserSettings.InitialHeight, null);
		yPos += spacing;

		Text t = profile.addText(xPos, yPos, "@ItemTemplate", null);
		t.setColor(Color.black);

		yPos += spacing;
		t = profile.addText(xPos, yPos, "@AnnotationTemplate", null);
		t.setColor(Color.magenta);

		yPos += spacing;
		t = profile.addText(xPos, yPos, "@CommentTemplate", null);
		t.setColor(Item.COLOR_WHEEL[3]);

		xPos = 600;
		yPos = intialYPos;

		Text help = profile.addText(xPos, yPos, "Expeditee Help", null);
		help.setSize(25);
		help.setFontStyle("Bold");
		help.setFamily("SansSerif");
		help.setColor(Item.COLOR_WHEEL[3]);

		xPos = 625;
		// Iterate through the help folder and add links
		File helpDirectory = new File(FrameIO.HELP_PATH);
		if (helpDirectory != null) {
			for (File helpFrameset : helpDirectory.listFiles()) {
				String framesetName = helpFrameset.getName();
				Frame indexFrame = FrameIO.LoadFrame(framesetName
						+ FrameIO.getFramesetExtension(framesetName));
				// Look through the folder for help index pages
				if (indexFrame != null
						&& ItemUtils.FindTag(indexFrame.getItems(),
								"@HelpIndex") != null) {
					yPos += spacing;
					t = profile.addText(xPos, yPos, indexFrame
							.getFramesetName(), null);
					t.setLink(indexFrame.getFrameName());
				}
			}
		}

		// FrameUtils.Parse(profile);
		FrameIO.SaveFrame(profile);
	}
}
