package org.expeditee.io;

import java.awt.Color;
import java.awt.Font;
import java.awt.Point;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;

import org.expeditee.gui.Frame;
import org.expeditee.items.Constraint;
import org.expeditee.items.Dot;
import org.expeditee.items.Item;
import org.expeditee.items.Line;
import org.expeditee.items.Text;

/**
 * Reads in Exp format files and constructs the Frame and Item objects they
 * contain.
 * 
 * @author mrww1
 * 
 */
public class ExpReader implements FrameReader {

	public static final String EXTENTION = ".exp";

	private BufferedReader _reader = null;

	private static LinkedHashMap<Character, Method> _ItemTags = null;

	private static LinkedHashMap<Character, Method> _FrameTags = null;

	private String _frameName;
	
	/**
	 * Does nothing, location must be set before use
	 */
	public ExpReader(String frameName) {
		_frameName = frameName;
		
		if (_ItemTags != null && _FrameTags != null)
			return;
		
		_ItemTags = new LinkedHashMap<Character, Method>();
		_FrameTags = new LinkedHashMap<Character, Method>();

		Class[] pString = { String.class };
		Class[] pInt = { int.class };
		Class[] pFloat = { float.class };
		Class[] pColor = { Color.class };
		Class[] pBool = { boolean.class };
		Class[] pFont = { Font.class };
		Class[] pPoint = { Point.class };
		Class[] pArrow = { int.class, double.class };
		Class[] pList = { List.class };
		Class[] pIntArray = { int[].class };
		Class[] pItem = { Item.class };

		try {
			_FrameTags.put('V', Frame.class.getMethod("setVersion", pInt));
			_FrameTags
					.put('p', Frame.class.getMethod("setProtection", pString));
			_FrameTags.put('U', Frame.class.getMethod("setOwner", pString));
			_FrameTags.put('D', Frame.class
					.getMethod("setDateCreated", pString));
			_FrameTags.put('M', Frame.class.getMethod("setLastModifyUser",
					pString));
			_FrameTags.put('d', Frame.class.getMethod("setLastModifyDate",
					pString));
			_FrameTags
					.put('F', Frame.class.getMethod("setFrozenDate", pString));

			_FrameTags.put('O', Frame.class.getMethod("setForegroundColor",
					pColor));
			_FrameTags.put('B', Frame.class.getMethod("setBackgroundColor",
					pColor));

			_ItemTags.put('S', Item.class.getMethod("setID", pInt));
			_ItemTags.put('s', Item.class.getMethod("setDateCreated", pString));
			_ItemTags.put('d', Item.class.getMethod("setColor", pColor));
			_ItemTags.put('G', Item.class.getMethod("setBackgroundColor",
					pColor));
			_ItemTags.put('P', Item.class.getMethod("setPosition", pPoint));
			_ItemTags.put('F', Item.class.getMethod("setLink", pString));
			_ItemTags.put('X', Item.class.getMethod("setAction", pList));
			_ItemTags.put('x', Item.class.getMethod("setActionMark", pBool));
			_ItemTags.put('U', Item.class.getMethod("setActionCursorEnter",
					pList));
			_ItemTags.put('V', Item.class.getMethod("setActionCursorLeave",
					pList));
			_ItemTags.put('W', Item.class.getMethod("setActionEnterFrame",
					pList));
			_ItemTags.put('Y', Item.class.getMethod("setActionLeaveFrame",
					pList));
			_ItemTags.put('D', Item.class.getMethod("setData", pString));
			_ItemTags.put('u', Item.class.getMethod("setHighlight", pBool));
			_ItemTags.put('e', Item.class.getMethod("setFillColor", pColor));
			_ItemTags.put('i', Item.class.getMethod("setFillPattern", pString));
			_ItemTags.put('o', Item.class.getMethod("setOwner", pString));
			_ItemTags.put('n', Item.class.getMethod("setLinkMark", pBool));
			_ItemTags
					.put('q', Item.class.getMethod("setLinkFrameset", pString));
			_ItemTags
					.put('y', Item.class.getMethod("setLinkTemplate", pString));
			_ItemTags.put('g', Item.class
					.getMethod("setLinePattern", pIntArray));

			_ItemTags.put('j', Item.class.getMethod("setArrow", pArrow));

			_ItemTags.put('f', Text.class.getMethod("setFont", pFont));
			_ItemTags.put('t', Text.class.getMethod("setSpacing", pInt));
			_ItemTags.put('T', Text.class.getMethod("appendLine", pString));

			_ItemTags.put('a', Text.class.getMethod("setWordSpacing", pInt));
			_ItemTags.put('b', Text.class.getMethod("setLetterSpacing", pInt));
			_ItemTags.put('m', Text.class.getMethod("setInitialSpacing", pInt));
			_ItemTags.put('w', Text.class.getMethod("setWidth", pInt));
			_ItemTags.put('k', Text.class.getMethod("setJustification", pInt));

			_ItemTags.put('h', Dot.class.getMethod("setThickness", pFloat));
			_ItemTags.put('l', Item.class.getMethod("setLineIDs", pString));
			_ItemTags
					.put('c', Dot.class.getMethod("setConstraintIDs", pString));

			// Lines and constraints are created differently
			_ItemTags.put('L', Line.class.getMethod("setStartItem", pItem));
			_ItemTags.put('C', Constraint.class.getMethod("getID",
					(Class[]) null));
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Determines whether a string begins with a KMS tag.
	 * 
	 * @param s
	 *            a line of text
	 * @return true if s begins with a KMS tag
	 */
	private boolean isValidLine(String s) {
		return s.length() >= 2 && s.charAt(1) == ' '
				&& Character.isLetter(s.charAt(0));
	}

	/**
	 * Reads a KMS file with the given name from disk.
	 * 
	 * @param frameName
	 *            the name of the Frame to read in from a file.
	 * @return A new Frame object that contains any Items described in the file.
	 * @throws IOException
	 *             Any exceptions occured by the BufferedReader.
	 */
	public Frame readFrame(String fullpath) throws IOException {
		_reader = new BufferedReader(new FileReader(fullpath));
		String next = "";
		Frame newFrame = new Frame();

		try {
			//Framename must be set before setting the frame number
			newFrame.setFrameName(_frameName);
			
			// First read all the header lines
			while (_reader.ready() && !(next = _reader.readLine()).equals("Z")) {
				if (isValidLine(next)) {
					processHeaderLine(newFrame, next);
				}
			}
			
			// Now read all the items
			Item currentItem = null;
			while (_reader.ready() && !(next = _reader.readLine()).equals("Z")) {
				// if this is the start of a new item add a new item
				if (isValidLine(next) ) {
					if (getTag(next) == 'S') {
						String value = getValue(next);
						int id = Integer.parseInt(value.substring(2));
						
						switch (value.charAt(0)) {
						case 'P': //check if its a point
							currentItem = new Dot(id);
							_linePoints.put(currentItem.getID(), currentItem);
							break;
						default:
							currentItem = new Text(id);
							break;
						}
						newFrame.addItem(currentItem);
					} else if (currentItem != null) {
						processBodyLine(currentItem, next);
					}
				}
			}

			// Read the lines
			while (_reader.ready() && !(next = _reader.readLine()).equals("Z")) {
				if (isValidLine(next)) {
					java.awt.Point idtype = separateValues(next.substring(2));
					// The next line must be the endpoints
					if (!_reader.ready())
						throw new Exception("Unexpected end of file");
					next = _reader.readLine();
					java.awt.Point startend = separateValues(next.substring(2));
					int start = startend.x;
					int end = startend.y;

					if (_linePoints.get(start) != null
							&& _linePoints.get(end) != null) {
						newFrame.addItem(new Line(_linePoints.get(start),
								_linePoints.get(end), idtype.x));
					} else {
						System.out
								.println("Error reading line with unknown end points");
					}
				}
			}

			// Read the constraints
			while (_reader.ready() && !(next = _reader.readLine()).equals("Z")) {
				if (isValidLine(next)) {
					java.awt.Point idtype = separateValues(next.substring(2));
					// The next line must be the endpoints
					if (!_reader.ready())
						throw new Exception("Unexpected end of file");
					next = _reader.readLine();
					java.awt.Point startend = separateValues(next.substring(2));

					Dot a = (Dot) _linePoints.get(startend.x);
					Dot b = (Dot) _linePoints.get(startend.y);

					new Constraint(a, b, idtype.x, idtype.y);
				}
			}

		} catch (Exception e) {
			System.out.println("Error reading frame file line: " + next + " "
					+ e.getMessage());
		}

		_reader.close();
		newFrame.setChanged(false);

		return newFrame;
	}

	// Stores points used when constructing lines
	private HashMap<Integer, Item> _linePoints = new HashMap<Integer, Item>();

	/**
	 * Processes the body section of the Exp file, which contains all Items
	 * 
	 * @param frame
	 *            The Frame to add any created Items to.
	 * @param line
	 *            The line of text read in from the file to process.
	 */
	private void processBodyLine(Item item, String line) {
		// separate the tag from the value
		Character tag = getTag(line);
		String value = getValue(line);

		Method toRun = _ItemTags.get(tag);
		if (toRun == null)
			System.out.println("Error accessing tag method: " + tag);
		Object[] vals = Conversion.Convert(toRun, value);

		try {
			if (vals != null)
				toRun.invoke(item, vals);
		} catch (Exception e) {
			System.out.println("Error running tag method: " + tag);
			//e.printStackTrace();
		}
	}

	private Character getTag(String line) {
		assert (line.length() > 0);
		return line.charAt(0);
	}

	private String getValue(String line) {
		if (line.length() > 2)
			return line.substring(2);
		else
			return "";
	}

	/**
	 * Reads the header section of the file, which contains information about
	 * the Frame.
	 * 
	 * @param frame
	 *            The Frame to assign the read values to.
	 * @param line
	 *            The line from the file to process.
	 * @return False if the end of the header has been reached, True otherwise.
	 */
	private void processHeaderLine(Frame frame, String line) throws IOException {
		// first separate the tag from the text
		Character tag = getTag(line);
		String value = getValue(line);
		Method toRun = _FrameTags.get(tag);

		if (toRun == null) {
			System.out.println("Tag '" + tag + "' in '" + line
					+ "' is not supported.");
			return;
		}

		Object[] vals = Conversion.Convert(toRun, value);
		try {
			toRun.invoke(frame, vals);
		} catch (Exception e) {
			System.out.println("Error running method: " + toRun.toGenericString());
			e.printStackTrace();
		}
	}

	// Returns a point from a String containing two ints separated by a space
	private java.awt.Point separateValues(String line) {
		int x = Integer.parseInt(line.substring(0, line.indexOf(" ")));
		int y = Integer.parseInt(line.substring(line.indexOf(" ") + 1));

		return new java.awt.Point(x, y);
	}
}
