package org.expeditee.items;

import java.awt.Graphics2D;
import java.awt.Polygon;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

/**
 * Represents a point on the screen. All Point objects are stored as x,y pairs
 * with an Item ID. Note: Lines and all combinates of lines (rectangles, etc)
 * are represented only as Points that share line IDs.
 * 
 * @author jdm18
 * 
 */
public class Dot extends Item {

	// Standard Item variables

	// contains all dots (including this one) that form an enclosure
	// if this dot is part of an enclosing shape
	private List<Dot> _enclosure = null;

	// the polygon surrounding this point, used for 'gravity'
	private Polygon _poly = null;

	// The thickness (in pixels) of this Point
	private float _thickness = -1;

	private int _pointType = Item.POINTTYPE_SQUARE;

	private boolean _filled = true;

	public Dot(int id) {
		super();
		setID(id);
	}

	/**
	 * Constructs a new Point with the given x,y coordinates and Item ID.
	 * 
	 * @param x
	 *            The x coordinate of this point
	 * @param y
	 *            The y coordinate of this point
	 * @param id
	 *            The Item ID of this Point
	 */
	public Dot(int x, int y, int id) {
		super();
		setID(id);
		setPosition(x, y);
	}

	public void setPointType(int type) {
		_pointType = type;
	}

	public void useFilledPoints(boolean val) {
		_filled = val;
	}

	@Override
	protected boolean hasVisibleArrow() {
		return getArrowheadRatio() != 0 && getArrowheadLength() != 0;
	}

	/**
	 * Sets the 'thickness' of this Dot, that is, how many pixels this Dot
	 * should be.
	 * 
	 * @param thick
	 *            The number of pixels to use when displaying this Dot The Dot
	 *            is displayed as a thick x thick square.
	 */
	@Override
	public void setThickness(float thick) {
		_thickness = thick;

		updatePolygon();

		// update the size of any lines
		for (Line line : getLines())
			line.setThickness(thick);
	}

	/**
	 * Returns the thickness (in pixels) of this Dot.
	 * 
	 * @return The 'thickness' of this Dot. (returns -1 if the thickness is not
	 *         set).
	 */
	@Override
	public float getThickness() {
		return _thickness;
	}

	public void paintFill(Graphics2D g) {
		if (getFillColor() != null && getEnclosingDots() != null) {
			g.setColor(getFillColor());
			g.fillPolygon(getEnclosedShape());
		}
	}

	@Override
	public void paint(Graphics2D g) {
		if (isHighlighted()) {
			g.setColor(getHighlightColor());
			g.drawPolygon(getPolygon());
		}

		// dots on lines are hidden
		if (getLines().size() > 0)
			return;

		g.setColor(getPaintColor());

		int thick = (int) _thickness;
		if (thick < 2)
			thick = 2;

		int width = thick / 2;

		// if(_thickness < 1)
		/*
		 * if(_filled) g.fillRect(getX(), getY(), thick,thick); else
		 * g.drawRect(getX(), getY(), thick,thick); else
		 */

		if (_pointType == Item.POINTTYPE_CIRCLE) {
			int points = 40;

			double radians = 0.0;
			int xPoints[] = new int[points];
			int yPoints[] = new int[xPoints.length];

			for (int i = 0; i < xPoints.length; i++) {
				xPoints[i] = getX() + (int) (width * Math.cos(radians));
				yPoints[i] = getY() + (int) (width * Math.sin(radians));
				radians += (2.0 * Math.PI) / xPoints.length;
			}

			if (_filled)
				g.fillPolygon(new Polygon(xPoints, yPoints, xPoints.length));
			else
				g.drawPolygon(new Polygon(xPoints, yPoints, xPoints.length));
		} else {
			if (_filled)
				g.fillRect(getX() - width, getY() - width, thick, thick);
			else
				g.drawRect(getX() - width, getY() - width, thick, thick);
		}

	}

	/**
	 * Updates the points of the polygon surrounding this Dot
	 */
	protected void updatePolygon() {
		int thick = (int) Math.ceil(getThickness());
		// Sets a minimum size for the dot
		thick = Math.max(thick, 10);

		int x = getX() - thick / 2;
		int y = getY() - thick / 2;

		_poly = new Polygon();
		_poly.addPoint(x - getGravity(), y - getGravity());
		_poly.addPoint(x + thick + getGravity(), y - getGravity());
		_poly.addPoint(x + thick + getGravity(), y + thick + getGravity());
		_poly.addPoint(x - getGravity(), y + thick + getGravity());
	}

	@Override
	public Polygon getPolygon() {
		if (_poly == null)
			updatePolygon();

		Polygon external = new Polygon(_poly.xpoints, _poly.ypoints,
				_poly.npoints);

		return external;
	}

	@Override
	public Dot copy() {
		Dot copy = new Dot(getX(), getY(), getID());

		Item.DuplicateItem(this, copy);

		copy.setThickness(getThickness());
		copy.setPointType(_pointType);
		copy.useFilledPoints(_filled);

		return copy;
	}

	@Override
	public int showHighlight(boolean val) {
		super.showHighlight(val);

		return Item.DEFAULT_CURSOR;
	}

	@Override
	public int getSize() {
		return (int) getThickness();
	}

	@Override
	public void setSize(int size) {
	}

	@Override
	public void setAnnotation(boolean val) {
	}

	@Override
	public boolean isAnnotation() {
		return false;
	}

	@Override
	public Item merge(Item merger, int mouseX, int mouseY) {
		// if the item being merged with is another Dot
		if (merger instanceof Dot) {
			Dot dot = (Dot) merger;

			// prevent concurrency issues if removing lines
			List<Line> lines = new ArrayList<Line>();
			lines.addAll(dot.getLines());

			for (Line line : lines) {
				// remove lines that are in common
				if (getLines().contains(line)) {
					dot.removeLine(line);
					removeLine(line);
				} else {
					// check for duplicated lines as a result of merging
					Dot check = (Dot) line.getOppositeEnd(dot);
					boolean dup = false;

					for (Line l : getLines()) {
						Item opposite = l.getOppositeEnd(this);

						if (check == opposite) {
							line.getStartItem().removeLine(line);
							line.getEndItem().removeLine(line);
							dup = true;
							break;
						}
					}

					if (!dup)
						line.replaceEnd(dot, this);
				}
			}

			setThickness(dot.getThickness());
			setColor(dot.getColor());
			setFillColor(dot.getFillColor());

			return null;
		}

		if (merger instanceof Text) {
			List<Line> lines = new LinkedList<Line>();
			lines.addAll(getLines());
			for (Line line : lines)
				line.replaceEnd(this, merger);

			removeAllLines();
			return merger;
		}

		// if the item being merged with is a Line
		if (merger instanceof Line) {
			Line line = (Line) merger;
			// if this dot is part of the line then keep it, otherwise it
			// can be removed
			if (line.getOppositeEnd(this) != null && getLines().contains(line))
				return merger;
			else
				return null;
		}

		return merger;
	}

	/**
	 * Sets the list of Dots (including this one) that form a closed shape.
	 * Passing null sets this dot back to its normal (non-enclosed) state.
	 * 
	 * @param enclosed
	 *            The List of Dots including this one that form a closed shape,
	 *            or null.
	 */
	public void setEnclosedList(List<Dot> enclosed) {
		_enclosure = enclosed;
	}

	/**
	 * Returns the polygon that represents the shape created by all the Dots in
	 * this Dot's enclosed list. If the list is null, then null is returned.
	 * 
	 * @return A Polygon the same shape and position as created by the Dots in
	 *         the enclosed list.
	 */
	public Polygon getEnclosedShape() {
		if (_enclosure == null)
			return null;

		Polygon poly = new Polygon();
		for (Dot d : _enclosure) {
			poly.addPoint(d.getX(), d.getY());
		}

		return poly;
	}

	/**
	 * Returns the list of Dots that, along with this Dot, form an enclosed
	 * polygon. If this Dot is not part of an enclosure null may be returned.
	 * 
	 * @return The List of Dots that form an enclosed shape with this Dot, or
	 *         null if this Dot is not part of an enclosure.
	 */
	public List<Dot> getEnclosingDots() {
		return _enclosure;
	}

	/**
	 * Returns whether this Dot has an assigned enclosure list of other Dots.
	 * The result is the same as getEnclosedShape() != null.
	 * 
	 * @return True if this Dot has an enclosure list of other Dots, false
	 *         otherwise.
	 */
	public boolean isEnclosed() {
		return _enclosure != null;
	}

	@Override
	public String getTypeAndID() {
		return "P " + getID();
	}
}
