package org.expeditee.items;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Point;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.image.ImageObserver;
import java.io.File;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.List;

import javax.imageio.ImageIO;
import javax.swing.ImageIcon;

import org.expeditee.gui.Frame;
import org.expeditee.gui.FrameGraphics;
import org.expeditee.gui.FrameIO;
import org.expeditee.io.Logger;

/**
 * This class represents an Image loaded from a file which is shown on the
 * screen. Loading of the Image from disk occurs in the constructor, and takes
 * approximately one second per mb of the Image file size. <br>
 * <br>
 * Currently Supported (Tested) Image formats:<br>
 * BMP<br>
 * JPG<br>
 * GIF<br>
 * GIF (Animated)<br>
 * <br>
 * Currently only the default size of the Image is supported, but future
 * versions may support scaling.
 * 
 * @author jdm18
 * 
 */
public class Picture extends Item {

	public static final int WIDTH = 0;

	public static final int RATIO = 1;

	public static final int FRACTION = 2;

	private Image _image = null;

	private Polygon _poly = null;

	// private Polygon _circle = null;
	private Text _source = null;

	// true if this is a frame picture
	private boolean _isFrame = false;

	private int _numerator = -1;

	private int _denominator = -1;

	private float _ratio = 1.0f;

	private int _width = -1;

	private int _scaleType = RATIO;

	private float _scale = 1.0f;

	private Point _cropStart = null;

	private Point _cropEnd = null;

	private Point _cropOrigin = null;

	private boolean _showCropping = false;

	// used to repaint animated GIF images, among other things.
	private ImageObserver _imageObserver = null;

	/**
	 * Creates a new Picture from the given path. The ImageObserver is optional
	 * and can be set to NULL. <br>
	 * Note: It is assumed that the file described in path has already been
	 * checked to exist.
	 * 
	 * @param source
	 *            The Text Item that was used to create this Picture
	 * @param path
	 *            The Path of the Image to load from disk.
	 * @param observer
	 *            The ImageObserver to assign when painting the Image on the
	 *            screen.
	 */
	public Picture(Text source, String path, String size, ImageObserver observer)
			throws IllegalArgumentException {
		super();
		_source = source;
		_imageObserver = observer;

		// Check if the picture is being created with @f
		if (path == null && source.getLink() != null) {
			String link = source.getLink();
			if (FrameIO.isPositiveInteger(link)){
				link = source.getParent().getFramesetNameAdjusted() + link;
			}
			Frame frame = FrameIO.LoadFrame(link);
			FrameGraphics.UpdateBuffer(frame, false);
			_image = frame.getBuffer();
			_isFrame = true;

		} else {

			// ImageIcon is faster, but cannot handle some formats
			// (notably
			// .bmp)
			// hence, we try this first, then if it fails we try
			// ImageIO
			_image = new ImageIcon(path).getImage();

			// if ImageIcon failed to read the image
			if (_image.getWidth(null) <= 0) {
				try {
					_image = ImageIO.read(new File(path));
				} catch (IOException e) {
					// e.printStackTrace();
					Logger.Log(e);
					_image = null;
				}
			}
		}

		if (_image == null)
			throw new IllegalArgumentException();

		try {
			if (size.length() == 0) {
				size = "" + _image.getWidth(null);
				_source.setText(_source.getFirstLine() + " " + size);
			}

			// parse size from text
			if (size.contains("/")) {
				// this is a fraction
				_numerator = Integer.parseInt(size.substring(0, size
						.indexOf("/")));
				_denominator = Integer.parseInt(size.substring(size
						.indexOf("/") + 1));
				_scale = (_numerator * 1.0f) / _denominator;
				_scaleType = FRACTION;
			} else if (size.contains(".")) {
				// this is a ratio
				_ratio = Float.parseFloat(size);
				_scaleType = RATIO;
				_scale = _ratio;
			} else if (size.length() > 0) {
				// this is an absolute width
				_width = Integer.parseInt(size);
				_scaleType = WIDTH;
				_scale = _width / (_image.getWidth(null) * 1.0f);
			}
		} catch (Exception e) {
			size = "" + _image.getWidth(null);
			_source.setText("@f " + size);
			FrameGraphics.ErrorMessage("Invalid argument for @f tag");
		}

	}

	public void setStartCrop(int x, int y) {
		_cropStart = new Point(x - getX(), y - getY());
		_cropOrigin = new Point(getX(), getY());
	}

	public void setEndCrop(int x, int y) {
		_cropEnd = new Point(x - getX(), y - getY());
	}

	public void setShowCrop(boolean value) {
		_showCropping = value;
	}

	public int getCroppedSize() {
		if (_cropStart == null || _cropEnd == null)
			return 0;

		int diff = (_cropEnd.x - _cropStart.x);
		diff *= (_cropEnd.y - _cropStart.y);
		return diff;
	}

	public void clearCropping() {
		_cropStart = null;
		_cropEnd = null;
		setShowCrop(false);
	}

	protected void updatePolygon() {
		if (_image == null)
			return;

		_poly = new Polygon();

		if (_cropStart == null || _cropEnd == null) {
			int width = _image.getWidth(null);
			int height = _image.getHeight(null);

			width *= _scale;
			height *= _scale;

			int xdiff = 0;

			if (getLink() != null || getAction() != null) {
				xdiff -= 10;
			}

			// extra pixel around the image so the highlighting is visible
			_poly.addPoint(_source.getX() - 1 + xdiff, _source.getY() - 1);
			_poly.addPoint(_source.getX() + 1 + width, _source.getY() - 1);
			_poly.addPoint(_source.getX() + 1 + width, _source.getY() + 1
					+ height);
			_poly.addPoint(_source.getX() - 1 + xdiff, _source.getY() + 1
					+ height);
		} else {
			Rectangle clip = new Rectangle(_source.getX() + _cropStart.x
					+ _cropOrigin.x, _source.getY() + _cropStart.y
					+ _cropOrigin.y, _cropEnd.x - _cropStart.x, _cropEnd.y
					- _cropStart.y).getBounds();
			_poly.addPoint((int) clip.getMinX() - 1, (int) clip.getMinY() - 1);
			_poly.addPoint((int) clip.getMinX() - 1, (int) clip.getMaxY() + 1);
			_poly.addPoint((int) clip.getMaxX() + 1, (int) clip.getMaxY() + 1);
			_poly.addPoint((int) clip.getMaxX() + 1, (int) clip.getMinY() - 1);

		}
	}

	/**
	 * Returns the Text Item that was used to create this Picture object. This
	 * is required by KMSWriter for saving the Frame.
	 * 
	 * @return The Text Item used to create this Picture.
	 */
	public Text getText() {
		return _source;
	}

	@Override
	public int getID() {
		return _source.getID();
	}

	@Override
	public void setID(int newID) {
		_source.setID(newID);
	}

	@Override
	public int getY() {
		return _source.getY();
	}

	@Override
	public int getX() {
		return _source.getX();
	}

	@Override
	public void setPosition(int x, int y) {
		_source.setPosition(x, y);

		updatePolygon();
	}

	@Override
	public String getLink() {
		if (_source == null)
			return null;
		else
			return _source.getLink();
	}

	@Override
	public void setLink(String frameName) {
		if (_source == null)
			return;
		else {
			_source.setLink(frameName);

			if (_isFrame) {
				Frame frame = FrameIO.LoadFrame(_source.getLink());
				// if (frame.getBuffer() == null)
				FrameGraphics.UpdateBuffer(frame, false);
				_image = frame.getBuffer();
			}
		}

		updatePolygon();
	}

	@Override
	public void setAction(List<String> action) {
		if (_source == null)
			return;
		else
			_source.setAction(action);

		updatePolygon();
	}

	@Override
	public List<String> getAction() {
		if (_source == null)
			return null;
		else
			return _source.getAction();
	}

	@Override
	public void paint(Graphics2D g) {
		if (_image == null)
			return;

		int width = _image.getWidth(null);
		int height = _image.getHeight(null);

		width *= _scale;
		height *= _scale;

		if (getLink() != null || getAction() != null) {
			g.setColor(getPaintHighlightColor());

			AffineTransform at = new AffineTransform();
			AffineTransform orig = g.getTransform();
			at.translate(getX() - 12, getY() + (getBoundsHeight() / 2));
			g.setTransform(at);

			if (getLinkMark() && getLink() != null) {
				g.drawPolygon(getCircle());

				// if the link is not valid, cross out the circle
				if (!isLinkValid())
					g.drawPolygon(getCircleCross());
			}

			if (getActionMark() && getAction() != null) {
				g.drawPolygon(getCircle());
				g.fillPolygon(getCircle());

				// if the link is not valid, cross out the circle
				if (getLink() != null && !isLinkValid()) {
					g.setColor(getPaintBackgroundColor());
					g.drawPolygon(getCircleCross());
				}
			}
			g.setTransform(orig);
		}

		if (isHighlighted()) {
			g.setColor(getHighlightColor());
			g.drawPolygon(getPolygon());
		}

		// if we are showing the cropping, then show the original as transparent
		if (_showCropping && _cropStart != null && _cropEnd != null) {
			// show the full image as transparent
			float alpha = .5f;
			g.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER,
					alpha));
			g.drawImage(_image, _source.getX(), _source.getY(), width, height,
					_imageObserver);

			g.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER,
					1.0f));
			// show the cropped area normally
			Shape clip = new Rectangle(_source.getX() + _cropStart.x, _source
					.getY()
					+ _cropStart.y, _cropEnd.x - _cropStart.x, _cropEnd.y
					- _cropStart.y);
			g.setColor(getPaintHighlightColor());
			g.draw(clip);
			g.setClip(clip);

			g.drawImage(_image, _source.getX(), _source.getY(), width, height,
					_imageObserver);
			g.draw(clip);
			// if the image is cropped, but we are not showing the cropping
		} else if (_cropStart != null && _cropEnd != null) {
			Shape clip = new Rectangle(_source.getX() + _cropStart.x
					+ _cropOrigin.x, _source.getY() + _cropStart.y
					+ _cropOrigin.y, _cropEnd.x - _cropStart.x, _cropEnd.y
					- _cropStart.y);
			g.setClip(clip);

			g.drawImage(_image, _source.getX(), _source.getY(), width, height,
					_imageObserver);

			// otherwise, paint normally
		} else {
			g.drawImage(_image, _source.getX(), _source.getY(), width, height,
					_imageObserver);
		}
	}

	@Override
	public Polygon getPolygon() {
		if (_poly == null)
			updatePolygon();

		Polygon external = new Polygon(_poly.xpoints, _poly.ypoints,
				_poly.npoints);
		return external;
	}

	@Override
	public int showHighlight(boolean val) {
		super.showHighlight(val);

		return Item.DEFAULT_CURSOR;
	}

	@Override
	public Picture copy() {
		Picture p;

		if (_isFrame){
			p = ItemUtils.CreateFramePicture((Text) _source.copy(),
					_imageObserver);
		}else
			p = ItemUtils.CreatePicture((Text) _source.copy(), _imageObserver);

		Item.DuplicateItem(this, p);

		if (_cropStart != null)
			p.setStartCrop(_cropStart.x, _cropStart.y);

		if (_cropEnd != null)
			p.setEndCrop(_cropEnd.x, _cropEnd.y);

		return p;
	}

	@Override
	public void setMaxSize(Dimension d) {
		// TODO Auto-generated method stub
	}

	@Override
	public Color getColor() {
		return _source.getColor();
	}

	@Override
	public void setColor(Color c) {
		_source.setColor(c);
	}

	@Override
	public int getSize() {
		return _source.getSize();
	}

	@Override
	public void setSize(int size) {
		int diff = size - _source.getSize();
		float oldScale = _scale;

		int oldDenom = _denominator;
		int oldNumer = _numerator;
		float oldRatio = _ratio;
		int oldWidth = _width;

		switch (_scaleType) {
		case (FRACTION):
			_numerator += diff;
			_scale = (_numerator * 1.0f) / _denominator;
			break;
		case (RATIO):
			_ratio += (diff) * 0.05;
			_scale = _ratio;
			break;
		case (WIDTH):
			_width += 10 * diff;
			_scale = _width / (_image.getWidth(null) * 1.0f);
			break;
		}

		// picture must still be at least 10 pixels wide
		if ((_scale * _image.getWidth(null)) <= 10) {
			_scale = oldScale;

			_denominator = oldDenom;
			_numerator = oldNumer;
			_ratio = oldRatio;
			_width = oldWidth;
		}

		// update source text item
		String line = _source.getFirstLine();

		switch (_scaleType) {
		case (FRACTION):
			line = line.replace(oldNumer + "/" + oldDenom, _numerator + "/"
					+ _denominator);
			break;
		case (RATIO):
			DecimalFormat format = new DecimalFormat("0.00");
			line = line.replace(format.format(oldRatio), format.format(_ratio));
			break;
		case (WIDTH):
			line = line.replace("" + oldWidth, "" + _width);
			break;
		}

		_source.setText(line);
		updatePolygon();
	}

	@Override
	public void setAnnotation(boolean val) {
		// TODO Auto-generated method stub

	}

	@Override
	public boolean isAnnotation() {
		// TODO Auto-generated method stub
		return false;
	}

	/**
	 * Returns the Image that this Picture object is painting on the screen.
	 * This is used by Frame to repaint animated GIFs.
	 * 
	 * @return The Image that this Picture object represents.
	 */
	public Image getImage() {
		return _image;
	}

	@Override
	public Item merge(Item merger, int mouseX, int mouseY) {
		return merger;
	}

}
